from argparse import ArgumentParser, Namespace, SUPPRESS
import attr
import hashlib
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
from kgtk.value.kgtkvalue import KgtkValue

@attr.s(slots=True, frozen=True)
class KgtkIdBuilderOptions(KgtkFormat):
    DEFAULT_OVERWRITE: bool = False
    DEFAULT_VERIFY_ID_UNIQUE: bool = False

    # TODO: use an enum
    COMPACT_PREFIXED_STYLE: str = "compact-prefix" # XXX### where ### is in a higher radix
    CONCAT_NLN_STYLE: str = "node1-label-node2" # node1-label-node2
    CONCAT_NL_NUM_STYLE: str = "node1-label-num" # node1-label-#
    CONCAT_NLN_NUM_STYLE: str = "node1-label-node2-num" # node1-label-node2-#
    CONCAT_WITH_OLD_ID_STYLE: str = "node1-label-node2-id" # Tag on any existing ID value
    EMPTY_STYLE: str = "empty" # DO not generate an ID (useful for debugging and other special circumstances)
    PREFIXED_STYLE: str = "prefix###" # XXX###
    WIKIDATA_STYLE: str = "wikidata" # node1-label-node2/hash
    WIKIDATA_WITH_CLAIM_ID_STYLE: str = "wikidata-with-claim-id" # node1-label-node2/hash-claimidhash

    STYLES: typing.List[str] = [
        COMPACT_PREFIXED_STYLE,
        EMPTY_STYLE,
        CONCAT_NLN_STYLE,
        CONCAT_NL_NUM_STYLE,
        CONCAT_NLN_NUM_STYLE,
        CONCAT_WITH_OLD_ID_STYLE,
        PREFIXED_STYLE,
        WIKIDATA_STYLE,
        WIKIDATA_WITH_CLAIM_ID_STYLE,
    ]
    DEFAULT_STYLE: str = PREFIXED_STYLE
    DEFAULT_CONCAT_NUM_WIDTH: int = 4
    DEFAULT_VALUE_HASH_WIDTH: int = 6
    DEFAULT_CLAIM_ID_HASH_WIDTH: int = 8
    DEFAULT_CLAIM_ID_COLUMN_NAME: str = "claim_id"

    # Defaults for prefixed style IDs.
    DEFAULT_PREFIX: str = "E"
    DEFAULT_INITIAL_ID: int = 1
    DEFAULT_PREFIX_NUM_WIDTH: int = 1

    DEFAULT_ID_SEPARATOR: str = "-"

    old_id_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    new_id_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    overwrite_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_OVERWRITE)
    verify_id_unique: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_VERIFY_ID_UNIQUE)
    id_style: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_STYLE)
    id_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_PREFIX)
    initial_id: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_INITIAL_ID)
    id_prefix_num_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_PREFIX_NUM_WIDTH)
    id_concat_num_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_CONCAT_NUM_WIDTH)
    value_hash_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_VALUE_HASH_WIDTH)
    claim_id_hash_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_CLAIM_ID_HASH_WIDTH)
    claim_id_column_name: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_CLAIM_ID_COLUMN_NAME)
    id_separator: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_ID_SEPARATOR)

    @classmethod
    def add_arguments(cls,
                      parser: ArgumentParser,
                      expert: bool = False,
                      overwrite: typing.Optional[bool] = None,
                      default_style: str = DEFAULT_STYLE):

        # This helper function makes it easy to suppress options from
        # The help message.  The options are still there, and initialize
        # what they need to initialize.
        def h(msg: str)->str:
            if expert:
                return msg
            else:
                return SUPPRESS

        if overwrite is None:
            overwrite = cls.DEFAULT_OVERWRITE

        # This one is likely to cause conflicts in the future.
        #
        # The default value is indirect.
        parser.add_argument(      "--old-id-column-name", dest="old_id_column_name",
                                  metavar="COLUMN_NAME",
                                  help=h("The name of the old ID column. (default=id)."))
            
        parser.add_argument(      "--new-id-column-name", dest="new_id_column_name",
                                  metavar="COLUMN_NAME",
                                  help=h("The name of the new ID column. (default=id)."))
            
        parser.add_argument(      "--overwrite-id", dest="overwrite_id",
                                  metavar="optional true|false",
                                  help="When true, replace existing ID values.  When false, copy existing ID values. " +
                                  "When --overwrite-id is omitted, it defaults to %(default)s. " +
                                  "When --overwrite-id is supplied without an argument, it is %(const)s.",
                                  type=optional_bool, nargs='?', const=True, default=overwrite)

        parser.add_argument(      "--verify-id-unique", dest="verify_id_unique",
                                  metavar="optional true|false",
                                  help="When true, verify ID uniqueness using an in-memory set of IDs. " +
                                  "When --verify-id-unique is omitted, it defaults to %(default)s. " +
                                  "When --verify-id-unique is supplied without an argument, it is %(const)s. ",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_VERIFY_ID_UNIQUE)

        parser.add_argument(      "--id-style", dest="id_style", default=default_style, choices=cls.STYLES,
                                  help=h("The ID generation style. (default=%(default)s)."))

        parser.add_argument(      "--id-prefix", dest="id_prefix", default=cls.DEFAULT_PREFIX,
                                  metavar="PREFIX",
                                  help=h("The prefix for a prefix### ID. (default=%(default)s)."))

        parser.add_argument(      "--initial-id", dest="initial_id", type=int, default=cls.DEFAULT_INITIAL_ID,
                                  metavar="INTEGER",
                                  help=h("The initial numeric value for a prefix### ID. (default=%(default)s)."))

        parser.add_argument(      "--id-prefix-num-width", dest="id_prefix_num_width", type=int, default=cls.DEFAULT_PREFIX_NUM_WIDTH,
                                  metavar="INTEGER",
                                  help=h("The width of the numeric value for a prefix### ID. (default=%(default)s)."))

        parser.add_argument(      "--id-concat-num-width", dest="id_concat_num_width", type=int, default=cls.DEFAULT_CONCAT_NUM_WIDTH,
                                  metavar="INTEGER",
                                  help=h("The width of the numeric value for a concatenated ID. (default=%(default)s)."))

        parser.add_argument(      '--value-hash-width', action="store", type=int, dest="value_hash_width", default=cls.DEFAULT_VALUE_HASH_WIDTH,
                                  help='How many characters should be used in a value hash? (default=%(default)d)')

        parser.add_argument(     '--claim-id-hash-width', action="store", type=int, dest="claim_id_hash_width", default=cls.DEFAULT_CLAIM_ID_HASH_WIDTH,
                                 help='How many characters should be used to hash the claim ID? 0 means do not hash the claim ID. (default=%(default)d)')

        parser.add_argument(     '--claim-id-column-name', action="store", type=str, dest="claim_id_column_name", default=cls.DEFAULT_CLAIM_ID_COLUMN_NAME,
                                 help='The name of the claim_id column. (default=%(default)s)')

        parser.add_argument(     '--id-separator', action="store", type=str, dest="id_separator", default=cls.DEFAULT_ID_SEPARATOR,
                                 help='The separator user between ID subfields. (default=%(default)s)')


    @classmethod
    def from_dict(cls, d: dict)->'KgtkIdBuilderOptions':
        return cls(
            old_id_column_name=d.get("old_id_column_name"),
            new_id_column_name=d.get("new_id_column_name"),
            overwrite_id=d.get("overwrite_id", False),
            verify_id_unique=d.get("verify_id_unique", False),
            id_style=d.get("id_style", cls.DEFAULT_STYLE),
            id_prefix=d.get("id_prefix", cls.DEFAULT_PREFIX),
            initial_id=d.get("initial_id", cls.DEFAULT_INITIAL_ID),
            id_prefix_num_width=d.get("id_prefix_num_width", cls.DEFAULT_PREFIX_NUM_WIDTH),
            id_concat_num_width=d.get("id_concat_num_width", cls.DEFAULT_CONCAT_NUM_WIDTH),
            value_hash_width=d.get("value_hash_width", cls.DEFAULT_VALUE_HASH_WIDTH),
            claim_id_hash_width=d.get("claim_id_hash_width", cls.DEFAULT_CLAIM_ID_HASH_WIDTH),
            claim_id_column_name=d.get("claim_id_column_name", cls.DEFAULT_CLAIM_ID_COLUMN_NAME),
            id_separator=d.get("id_separator", cls.DEFAULT_ID_SEPARATOR),
        )

    # Build the value parsing option structure.
    @classmethod
    def from_args(cls, args: Namespace)->'KgtkIdBuilderOptions':
        return cls.from_dict(vars(args))

    def show(self, out: typing.TextIO=sys.stderr):
        if self.old_id_column_name is not None:
            print("--old-id-column-name=%s" % str(self.old_id_column_name), file=out, flush=True)
        if self.new_id_column_name is not None:
            print("--new-id-column-name=%s" % str(self.new_id_column_name), file=out, flush=True)
        print("--overwrite-id=%s" % str(self.overwrite_id), file=out, flush=True)
        print("--verify-id-unique=%s" % str(self.verify_id_unique), file=out, flush=True)
        print("--id-style=%s" % str(self.id_style), file=out, flush=True)
        print("--id-prefix=%s" % repr(self.id_prefix), file=out, flush=True)
        print("--initial-id=%s" % str(self.initial_id), file=out, flush=True)
        print("--id-prefix-num-width=%s" % str(self.id_prefix_num_width), file=out, flush=True)
        print("--id-concat-num-width=%s" % str(self.id_concat_num_width), file=out, flush=True)
        print("--value-hash-width=%s" % str(self.value_hash_width), file=out, flush=True)
        print("--claim-id-hash-width=%s" % str(self.claim_id_hash_width), file=out, flush=True)
        print("--claim-id-column-name=%s" % repr(self.claim_id_column_name), file=out, flush=True)
        print("--id-separator=%s" % repr(self.id_separator), file=out, flush=True)


@attr.s(slots=True, frozen=False)
class KgtkIdBuilder(KgtkFormat):
    options: KgtkIdBuilderOptions = attr.ib(validator=attr.validators.instance_of(KgtkIdBuilderOptions))

    column_names: typing.List[str]= attr.ib() # TODO: add validator

    old_id_column_name: str = attr.ib(validator=attr.validators.instance_of(str))
    old_id_column_idx: int = attr.ib(validator=attr.validators.instance_of(int))

    new_id_column_name: str = attr.ib(validator=attr.validators.instance_of(str))
    new_id_column_idx: int = attr.ib(validator=attr.validators.instance_of(int))
    add_new_id_column: bool = attr.ib(validator=attr.validators.instance_of(bool))

    node1_column_idx: int = attr.ib(validator=attr.validators.instance_of(int))
    label_column_idx: int= attr.ib(validator=attr.validators.instance_of(int))
    node2_column_idx: int= attr.ib(validator=attr.validators.instance_of(int))

    claim_id_column_idx: int= attr.ib(validator=attr.validators.instance_of(int))

    current_id: int = attr.ib(validator=attr.validators.instance_of(int))

    id_set: typing.Set[str] =attr.ib(validator=attr.validators.instance_of(set), factory=set)

    nl_keys: typing.MutableMapping[str, int] = attr.ib(validator=attr.validators.instance_of(dict), factory=dict)

    @classmethod
    def new(cls,
            kr: KgtkReader,
            options: KgtkIdBuilderOptions,
    )->'KgtkIdBuilder':
        """
        This is the KgtkIdBuilder factory method.
        """
        column_names: typing.List[str] = kr.column_names.copy()
        column_name_map: typing.Mapping[str, int] = kr.column_name_map
        node1_column_idx: int = kr.node1_column_idx
        label_column_idx: int = kr.label_column_idx
        node2_column_idx: int = kr.node2_column_idx
        id_column_idx: int = kr.id_column_idx

        return cls.new1(column_names, column_name_map, node1_column_idx, label_column_idx, node2_column_idx, id_column_idx, options)

    @classmethod
    def from_column_names(cls,
             column_names: typing.List[str],
             options: KgtkIdBuilderOptions,
    )->'KgtkIdBuilder':
        column_name_map: typing.MutableMapping[str, int] = { }
        column_idx: int
        column_name: str
        for column_idx, column_name in enumerate(column_names):
            column_name_map[column_name] = column_idx
        
        node1_column_idx: int
        label_column_idx: int
        node2_column_idx: int
        id_column_idx: int

        node1_column_idx, label_column_idx, node2_column_idx, id_column_idx = \
            KgtkReader.get_special_columns(column_name_map, "", "idbuilder")

        return cls.new1(column_names.copy(), column_name_map, node1_column_idx, label_column_idx, node2_column_idx, id_column_idx, options)

    @classmethod
    def new1(cls,
             column_names: typing.List[str],
             column_name_map: typing.Mapping[str, int],
             node1_column_idx: int,
             label_column_idx: int,
             node2_column_idx: int,
             id_column_idx: int,
             options: KgtkIdBuilderOptions,
    )->'KgtkIdBuilder':

        # Find the old ID column.
        old_id_column_idx: int
        old_id_column_name: str
        if options.old_id_column_name is not None:
            # The old ID column was explicitly named.
            old_id_column_name = options.old_id_column_name
            if old_id_column_name in column_name_map:
                # The new ID column already exists.
                old_id_column_idx = column_name_map[old_id_column_name]
            else:
                raise ValueError("The old ID column named '%s' is not known." % options.old_id_column_name)
        else:
            # The old ID column was not explicitly named.
            if id_column_idx >= 0:
                # The input file has an ID column.  Use it.
                old_id_column_name = column_names[id_column_idx]
                old_id_column_idx = id_column_idx
            else:
                # There is not old ID column index.
                old_id_column_idx = -1
                old_id_column_name = ""

        # Find the new ID column.
        new_id_column_idx: int
        add_new_id_column: bool
        new_id_column_name: str
        if options.new_id_column_name is not None:
            # The new ID column was explicitly named.
            new_id_column_name = options.new_id_column_name
            if new_id_column_name in column_name_map:
                # The new ID column already exists.
                new_id_column_idx = column_name_map[new_id_column_name]
                add_new_id_column = False
            else:
                # Create a new ID column.
                new_id_column_idx = len(column_names)
                column_names.append(new_id_column_name)
                add_new_id_column = True
        else:
            # The new ID column was not explicitly named.
            if id_column_idx >= 0:
                # The input file has an ID column.  Use it.
                new_id_column_name = column_names[id_column_idx]
                new_id_column_idx = id_column_idx
                add_new_id_column = False
            else:
                # Create a new ID column.
                new_id_column_idx = len(column_names)
                new_id_column_name = KgtkFormat.ID
                column_names.append(KgtkFormat.ID)
                add_new_id_column = True

        claim_id_column_idx: int = column_name_map.get(options.claim_id_column_name, -1)

        # Any prerequisites?
        if options.id_style == KgtkIdBuilderOptions.CONCAT_NLN_STYLE:
            if node1_column_idx < 0:
                raise ValueError("No node1 column index")
            if label_column_idx < 0:
                raise ValueError("No label column index")
            if node2_column_idx < 0:
                raise ValueError("No node2 column index")

        elif options.id_style == KgtkIdBuilderOptions.CONCAT_NL_NUM_STYLE:
            if node1_column_idx < 0:
                raise ValueError("No node1 column index")
            if label_column_idx < 0:
                raise ValueError("No label column index")

        elif options.id_style == KgtkIdBuilderOptions.CONCAT_NLN_NUM_STYLE:
            if node1_column_idx < 0:
                raise ValueError("No node1 column index")
            if label_column_idx < 0:
                raise ValueError("No label column index")
            if node2_column_idx < 0:
                raise ValueError("No node2 column index")

        elif options.id_style == KgtkIdBuilderOptions.CONCAT_WITH_OLD_ID_STYLE:
            if node1_column_idx < 0:
                raise ValueError("No node1 column index")
            if label_column_idx < 0:
                raise ValueError("No label column index")
            if node2_column_idx < 0:
                raise ValueError("No node2 column index")
            if old_id_column_idx < 0:
                raise ValueError("No old ID column index")

        elif options.id_style == KgtkIdBuilderOptions.WIKIDATA_STYLE:
            if node1_column_idx < 0:
                raise ValueError("No node1 column index")
            if label_column_idx < 0:
                raise ValueError("No label column index")
            if node2_column_idx < 0:
                raise ValueError("No node2 column index")
        
        elif options.id_style == KgtkIdBuilderOptions.WIKIDATA_WITH_CLAIM_ID_STYLE:
            if node1_column_idx < 0:
                raise ValueError("No node1 column index")
            if label_column_idx < 0:
                raise ValueError("No label column index")
            if node2_column_idx < 0:
                raise ValueError("No node2 column index")
            if claim_id_column_idx < 0:
                raise ValueError("No claim_id column index")
        
        return cls(options=options,
                   column_names=column_names,
                   old_id_column_name=old_id_column_name,
                   old_id_column_idx=old_id_column_idx,
                   new_id_column_name=new_id_column_name,
                   new_id_column_idx=new_id_column_idx,
                   add_new_id_column=add_new_id_column,
                   node1_column_idx=node1_column_idx,
                   label_column_idx=label_column_idx,
                   node2_column_idx=node2_column_idx,
                   claim_id_column_idx=claim_id_column_idx,
                   current_id=options.initial_id
        )

    def verify_uniqueness(self, id_value: str, row: typing.List[str], line_number, who: str):
        """
        Verify that ID values are not repeated.  This is OK for the output
        of `kgtk compact`, but is a little too strong for general use.
        The weaker constraint should be that the ID values don't repeat
        with different (node1, label, node2) tuples in an edge file.
        """
        if KgtkFormat.LIST_SEPARATOR in id_value:
            # The ID value might be a list.
            id_v: str
            for id_v in KgtkValue.split_list(id_value):
                if id_v in self.id_set:
                    # TODO: Probably want more error handling options, such as
                    # printing the offending row and choosing to continue.
                    raise ValueError("Line %d: %s ID '%s' duplicates a previous ID '%s'." % (line_number, who, id_value, id_v))
                else:
                    self.id_set.add(id_v)
        else:
            # Not a list, we can process this faster.
            if id_value in self.id_set:
                # TODO: Probably want more error handling options, such as
                # printing the offending row and choosing to continue.
                raise ValueError("Line %d: %s ID '%s' duplicates a previous ID '%s'." % (line_number, who, id_value, id_value))
            else:
                self.id_set.add(id_value)
        

    def build(self,
              row: typing.List[str],
              line_number: int,
              already_added: typing.Optional[bool]=False)->typing.List[str]:
        """
        Build a new ID value if needed.
        """
        row = list(row).copy() # as a precaution
        if self.add_new_id_column and not already_added:
            row.append("")
        elif self.old_id_column_idx >= 0:
            if row[self.old_id_column_idx] != "" and not self.options.overwrite_id:
                if self.new_id_column_idx != self.old_id_column_idx:
                    row[self.new_id_column_idx] = row[self.old_id_column_idx]
                if self.options.verify_id_unique:
                    self.verify_uniqueness(row[self.old_id_column_idx], row, line_number, "existing")

                return row

        # Perhaps we also want to lookup (node1, label, node2), and if an ID
        # has already been assigned, reuse it.

        new_id: str
        if self.options.id_style == KgtkIdBuilderOptions.COMPACT_PREFIXED_STYLE:
            new_id = self.build_compact_prefixed(row)
        elif self.options.id_style == KgtkIdBuilderOptions.CONCAT_NLN_STYLE:
            new_id = self.build_concat_nln(row)
        elif self.options.id_style == KgtkIdBuilderOptions.CONCAT_NL_NUM_STYLE:
            new_id = self.build_concat_nl_num(row)
        elif self.options.id_style == KgtkIdBuilderOptions.CONCAT_NLN_NUM_STYLE:
            new_id = self.build_concat_nln_num(row)
        elif self.options.id_style == KgtkIdBuilderOptions.CONCAT_WITH_OLD_ID_STYLE:
            new_id = self.build_concat_with_old_id(row)
        elif self.options.id_style == KgtkIdBuilderOptions.PREFIXED_STYLE:
            new_id = self.build_prefixed(row)
        elif self.options.id_style == KgtkIdBuilderOptions.EMPTY_STYLE:
            new_id = "" # Useful for debugging and other unusual circumstances.
        elif self.options.id_style == KgtkIdBuilderOptions.WIKIDATA_STYLE:
            new_id = self.build_wikidata_id(row)
        elif self.options.id_style == KgtkIdBuilderOptions.WIKIDATA_WITH_CLAIM_ID_STYLE:
            new_id = self.build_wikidata_id_with_claim_id(row)
        else:
            raise ValueError("Unknown ID style '%s'." % self.options.id_style)

        row[self.new_id_column_idx] = new_id
        if self.options.verify_id_unique:
            self.verify_uniqueness(new_id, row, line_number, "new")
        return row

    # Build more compact prefixed IDs by encoding the ID in a base that uses
    # the decimal digits and most of the lower and upper case letters.  Vowels
    # are omitted to avoid creating English language obscenities ('w' is retained,
    # which raises the issue of potential Welsh language obscenities).  The
    # resulting set of digits implies base 50.
    #
    # Note:  The base digits should start with '0' for zfill to be intuitive.
    # Conversely, '+' and '-' must not be in the base digits due to zfill.
    #
    # Note:  Negative ID values are supported, but they should probably be forbidden.
    #
    # TODO: Disallow negative ID values.
    base_digits: str = "0123456789bcdfghjklmnpqrstvwxzBCDFGHJKLMNPQRSTVWXZ"

    def build_compact_prefixed(self, row: typing.List[str])->str:
        id: int = self.current_id
        self.current_id += 1

        compact_id: str
        if id == 0:
            compact_id = self.base_digits[0]
        else:
            base: int = len(self.base_digits)
            encoded_digits: typing.List[str] = list()
            negative: bool = False
            if id < 0:
                negative = True
                id = - id
            while id:
                encoded_digits.append(self.base_digits[id % base])
                id //= base
            if negative:
                encoded_digits.append('-')
            encoded_digits.reverse()
            compact_id = ''.join(encoded_digits)
        
        return self.options.id_prefix + compact_id.zfill(self.options.id_prefix_num_width)

    def build_concat_nln(self, row: typing.List[str])->str:
        return row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + row[self.node2_column_idx]

    def build_concat_nl_num(self, row: typing.List[str])->str:
        # Returns node1-label-### where ### starts at 0.
        # Assumes that records are compact on (node2) or (node2, id), as needed.
        key: str = row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx]
        if key in self.nl_keys:
            self.nl_keys[key] += 1
        else:
            self.nl_keys[key] = 0
        return key + self.options.id_separator + str(self.nl_keys[key]).zfill(self.options.id_concat_num_width)

    def build_concat_nln_num(self, row: typing.List[str])->str:
        # Returns node1-label-node2-### where ### starts at 0.
        # Assumes that records are compact on (node2) or (node2, id), as needed.
        key: str = row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + row[self.node2_column_idx]
        if key in self.nl_keys:
            self.nl_keys[key] += 1
        else:
            self.nl_keys[key] = 0
        return key + self.options.id_separator + str(self.nl_keys[key]).zfill(self.options.id_concat_num_width)

    def build_concat_with_old_id(self, row: typing.List[str])->str:
        key: str = row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + row[self.node2_column_idx]
        if row[self.old_id_column_idx] != "":
            key += self.options.id_separator + row[self.old_id_column_idx]
        return key

    def build_prefixed(self, row: typing.List[str])->str:
        new_id: str =  self.options.id_prefix + str(self.current_id).zfill(self.options.id_prefix_num_width)
        self.current_id += 1
        return new_id

    def build_wikidata_id(self, row: typing.List[str])->str:
        node2_value: str = row[self.node2_column_idx]
        if self.options.value_hash_width > 0 and node2_value.startswith(('L', 'P', 'Q')):
            return self.build_concat_nln(row)
        else:
            return row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + \
                hashlib.sha256(node2_value.encode('utf-8')).hexdigest()[:self.options.value_hash_width]

    def build_wikidata_id_with_claim_id(self, row: typing.List[str])->str:
        claim_id: str = row[self.claim_id_column_idx]
        if len(claim_id) == 0:
            # TODO: Perhaps raise an exception?
            return self.build_wikidata_id(row)

        elif self.options.claim_id_hash_width == 0:
            return self.build_wikidata_id(row) + self.options.id_separator + claim_id.lower()

        else:
            return self.build_wikidata_id(row) + self.options.id_separator + hashlib.sha256(claim_id.lower().encode('utf-8')).hexdigest()[:self.options.claim_id_hash_width]

    def process(self, kr: KgtkReader, kw: KgtkWriter):
        line_number: int = 0
        row: typing.List[str]
        for row in kr:
            line_number += 1
            kw.write(self.build(row, line_number))

def main():
    """
    Test the KGTK ID builder.
    """
    parser: ArgumentParser = ArgumentParser()
    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data (default=%(default)s)", type=Path, nargs="?", default="-")
    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    KgtkIdBuilderOptions.add_arguments(parser, expert=True)
    
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_args(args)    
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

    # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("input: %s" % str(args.input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    # First create the KgtkReader.  It provides parameters used by the ID
    # column builder. Next, create the ID column builder, which provides a
    # possibly revised list of column names for the KgtkWriter.  Last, create
    # the KgtkWriter.

    # Open the input file.
    kr: KgtkReader =  KgtkReader.open(args.input_file_path,
                                      error_file=error_file,
                                      options=reader_options,
                                      value_options = value_options,
                                      verbose=args.verbose,
                                      very_verbose=args.very_verbose,
    )

    # Create the ID builder.
    idb: KgtkIdBuilder = KgtkIdBuilder.new(kr, idbuilder_options)

    # Open the output file.
    ew: KgtkWriter = KgtkWriter.open(idb.column_names,
                                     args.output_file_path,
                                     mode=kr.mode,
                                     require_all_columns=True,
                                     prohibit_extra_columns=True,
                                     fill_missing_columns=False,
                                     gzip_in_parallel=False,
                                     verbose=args.verbose,
                                     very_verbose=args.very_verbose)

    # Process the input file, building IDs.
    idb.process(kr, ew)

    ew.close()
    kr.close()

if __name__ == "__main__":
    main()



"""Add label columns for values in the node1, label, and node2 fields.

The input rows are saved in memory, as well as the value-to-label mapping.
This will impose a limit on the size of the input files that can be processed.

TODO: Take a list of properties?

TODO: Optionally save the input rows in an external disk file?

TODO: Optionally reread the input stream insted of saving the input rows?
      With special provision for copying standard input to an external disk file.

TODO: Need KgtkWriterOptions

TODO: Provide seperate reader options for the label file.

"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkLift(KgtkFormat):
    DEFAULT_OUTPUT_LIFTED_COLUMN_SEPARATOR: str = ";"
    DEFAULT_LABEL_SELECT_COLUMN_VALUE: str = "label"
    DEFAULT_OUTPUT_LIFTED_COLUMN_SUFFIX: str = DEFAULT_OUTPUT_LIFTED_COLUMN_SEPARATOR + DEFAULT_LABEL_SELECT_COLUMN_VALUE

    input_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    label_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
 
    input_select_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    input_select_column_value: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    input_lifting_column_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    output_select_column_value: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    output_lifted_column_suffix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_OUTPUT_LIFTED_COLUMN_SUFFIX)
    output_lifted_column_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    label_select_disable: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    label_select_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    label_select_column_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_LABEL_SELECT_COLUMN_VALUE)

    label_match_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    label_value_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    remove_label_records: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    suppress_duplicate_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    sort_lifted_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    suppress_empty_columns: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    ok_if_no_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    prefilter_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    input_is_presorted: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    labels_are_presorted: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    default_value: str = attr.ib(validator=attr.validators.instance_of(str), default="")

    clear_before_lift: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    overwrite: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)

    output_only_modified_rows: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    unmodified_row_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)), default=None)

    matched_label_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)), default=None)
    unmatched_label_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)), default=None)

    languages: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)
    prioritize: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)

    lift_all_columns: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    force_input_mode_none: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    
    # TODO: add rewind logic here and KgtkReader

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    input_reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    label_reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def build_lift_column_idxs(self, kr: KgtkReader)->typing.List[int]:
        lift_column_idxs: typing.List[int] = [ ]
        if self.input_lifting_column_names is not None and len(self.input_lifting_column_names) > 0:
            # Process a custom list of columns to be lifted.
            lift_column_name: str
            for lift_column_name in self.input_lifting_column_names:
                if lift_column_name not in kr.column_name_map:
                    raise ValueError("Unknown lift column %s." % lift_column_name)
                lift_column_idxs.append(kr.column_name_map[lift_column_name])

        elif self.lift_all_columns:
            idx: int
            column_name: str
            for idx, column_name in enumerate(kr.column_names):
                if not column_name.endswith(self.output_lifted_column_suffix):
                    lift_column_idxs.append(idx)

        else:
            # Use the edge file key columns if they exist.
            if kr.node1_column_idx >= 0:
                lift_column_idxs.append(kr.node1_column_idx)
            if kr.label_column_idx >= 0:
                lift_column_idxs.append(kr.label_column_idx)
            if kr.node2_column_idx >= 0:
                lift_column_idxs.append(kr.node2_column_idx)

        # Verify that we found some columns to lift:
        if len(lift_column_idxs) == 0:
            raise ValueError("No lift columns found.")

        return lift_column_idxs

    def lookup_input_select_column_idx(self, kr: KgtkReader)->int:
        input_select_column_idx: int
        if self.input_select_column_name is None:
            if kr.label_column_idx < 0:
                raise ValueError("No input select column index.")
            input_select_column_idx = kr.label_column_idx
        else:
            if self.input_select_column_name not in kr.column_name_map:
                raise ValueError("Input select column `%s` not found." % self.input_select_column_name)
            input_select_column_idx = kr.column_name_map[self.input_select_column_name]
        return input_select_column_idx

    def lookup_label_match_column_idx(self, kr: KgtkReader)->int:
        label_match_column_idx: int
        if self.label_match_column_name is None:
            if kr.is_edge_file:
                if self.verbose:
                    print("The label file is an edge file, defaulting to the node1 column (or alias) for the match column.",
                          file=self.error_file, flush=True)
                if kr.node1_column_idx < 0:
                    raise ValueError("Cannot find the label match column (node1) in the label file (an edge file).")
                label_match_column_idx = kr.node1_column_idx
            elif kr.is_node_file:
                if self.verbose:
                    print("The label file is a node file, defaulting to the ID column for the match column.",
                          file=self.error_file, flush=True)
                if kr.id_column_idx < 0:
                    raise ValueError("Cannot find the label match column (id) in the label file (a node file).")
                label_match_column_idx = kr.id_column_idx
            else:
                raise ValueError("No label match column specified and not an edge or node file.")

        else:
            if self.label_match_column_name not in kr.column_name_map:
                raise ValueError("Label match column `%s` not found." % self.label_match_column_name)
            label_match_column_idx = kr.column_name_map[self.label_match_column_name]
        return label_match_column_idx

    def lookup_label_select_column_idx(self, kr: KgtkReader)->int:
        label_select_column_idx: int
        if self.label_select_disable:
            label_select_column_idx = -1
        elif self.label_select_column_name is None:
            if kr.is_edge_file:
                if kr.label_column_idx < 0:
                    raise ValueError("Cannot find the label select column (label) in the label file.")
                label_select_column_idx = kr.label_column_idx
            elif kr.is_node_file:
                if self.verbose:
                    print("The label file is a node file, defaulting to no label select column.", file=self.error_file, flush=True)
                label_select_column_idx = -1 # Special case!
            else:
                raise ValueError("No label match column specified and not an edge or node file.")
        else:
            if self.label_select_column_name not in kr.column_name_map:
                raise ValueError("Label select column `%s` not found." % self.label_select_column_name)
            label_select_column_idx = kr.column_name_map[self.label_select_column_name]
        return label_select_column_idx

    def lookup_label_value_column_idx(self, kr: KgtkReader)->int:
        label_value_column_idx: int
        if self.label_value_column_name is None:
            if kr.is_edge_file:
                if self.verbose:
                    print("The label file is an edge file, defaulting to the node2 column (or alias) for the value column.",
                          file=self.error_file, flush=True)
                if kr.node2_column_idx < 0:
                    raise ValueError("Cannot find the label value column (node2) in the label edge file.")
                label_value_column_idx = kr.node2_column_idx
            elif kr.is_node_file:
                if self.verbose:
                    print("The label file is a node file, defaulting to the label column for the value column.",
                          file=self.error_file, flush=True)
                if kr.label_column_idx < 0:
                    raise ValueError("Cannot find the label value column (label) in the label node file.")
                label_value_column_idx = kr.label_column_idx
        else:
            if self.label_value_column_name not in kr.column_name_map:
                raise ValueError("Label value column `%s` not found in the label file." % self.label_value_column_name)
            label_value_column_idx = kr.column_name_map[self.label_value_column_name]

        return label_value_column_idx

    def lookup_label_table_idxs(self, kr: KgtkReader)->typing.Tuple[int, int, int]:
        label_match_column_idx: int = self.lookup_label_match_column_idx(kr)
        label_select_column_idx: int = self.lookup_label_select_column_idx(kr)
        label_value_column_idx: int = self.lookup_label_value_column_idx(kr)

        return label_match_column_idx, label_select_column_idx, label_value_column_idx

    def load_labels(self,
                    kr: KgtkReader,
                    path: Path,
                    save_input: bool = True,
                    labels_needed: typing.Optional[typing.Set[str]] = None,
                    label_rows: typing.Optional[typing.MutableMapping[str, typing.List[typing.List[str]]]] = None,
                    is_label_file: bool = False,
    )->typing.Tuple[typing.Mapping[str, str], typing.List[typing.List[str]]]:
        input_rows: typing.List[typing.List[str]] = [ ]
        labels: typing.MutableMapping[str, str] = { }

        label_match_column_idx: int
        label_select_column_idx: int
        label_value_column_idx: int
        label_match_column_idx, label_select_column_idx, label_value_column_idx = self.lookup_label_table_idxs(kr)

        # Build the label filter.  We will still do the same filtering steps in code below.
        #
        # TODO: Remove the redundant filtering steps.
        if is_label_file:
            label_filter: typing.MutableMapping[int, typing.Set[str]] = dict()
            if label_select_column_idx >= 0:
                label_filter[label_select_column_idx] = set([self.label_select_column_value])
            if labels_needed is not None and label_match_column_idx >= 0:
                label_filter[label_match_column_idx] = set(labels_needed)
            if len(label_filter) > 0:
                kr.add_input_filter(label_filter)

        if self.verbose:
            print("Loading labels from %s" % path, file=self.error_file, flush=True)
            if labels_needed is not None:
                print("Filtering for needed labels", file=self.error_file, flush=True)
            print("label_match_column_idx=%d (%s)." % (label_match_column_idx, kr.column_names[label_match_column_idx]), file=self.error_file, flush=True)
            if label_select_column_idx < 0:
                print("label_select_column_idx=%d." % (label_select_column_idx), file=self.error_file, flush=True)
            else:
                print("label_select_column_idx=%d (%s)." % (label_select_column_idx, kr.column_names[label_select_column_idx]), file=self.error_file, flush=True)
            print("label_value_column_idx=%d (%s)." % (label_value_column_idx, kr.column_names[label_value_column_idx]), file=self.error_file, flush=True)
            print("label_select_column_value='%s'." % self.label_select_column_value, file=self.error_file, flush=True)

        # Setup for language filtering:
        language_filtering: bool = self.languages is not None and len(self.languages) > 0
        allow_all_strings: bool = False
        allow_all_lqstrings: bool = False
        language_filter: typing.List[str] = list()
        language_filter_tuple: typing.Tuple[str, ...] = ()
        label_priorities: typing.MutableMapping[str, int] = dict()
        if language_filtering:
            for lf in self.languages:
                if lf == 'NONE':
                    if self.prioritize:
                        language_filter.append(lf)
                    else:
                        allow_all_strings = True
                elif lf == 'ANY':
                    if self.prioritize:
                        language_filter.append(lf)
                    else:
                        allow_all_lqstrings = True
                else:
                    language_filter.append("'@" + lf)
            language_filter_tuple = tuple(language_filter)

        key: str
        row: typing.List[str]
        for row in kr:
            if label_select_column_idx < 0 or row[label_select_column_idx] == self.label_select_column_value:
                # This is a label definition row.
                label_key: str = row[label_match_column_idx]
                if labels_needed is not None and label_key not in labels_needed:
                    # We don't need this label.
                    if save_input and not self.remove_label_records:
                        input_rows.append(row.copy())
                    continue # Ignore unneeded labels.

                label_value: str = row[label_value_column_idx]
                if label_value == self.default_value:
                    # We can ignore default values.
                    #
                    # TODO: There might be different semantics to default value handling between
                    # this code and the merge code.
                    if save_input and not self.remove_label_records:
                        input_rows.append(row.copy())
                    continue # Ignore default values.

                process_label: bool
                if language_filtering:
                    sigil: str
                    if len(label_value) == 0:
                        # An empty value is not a valid label when filtering by language.
                        process_label = False

                    elif self.prioritize:
                        current_priority: int = label_priorities.get(label_key, 999999)
                        sigil = label_value[0]
                        priority: int
                        lf: str
                        for priority, lf in enumerate(language_filter):
                            if priority >= current_priority:
                                process_label = False
                                break

                            elif sigil == KgtkFormat.STRING_SIGIL:
                                if lf == "NONE":
                                    process_label = True
                                    break
                                
                            elif sigil == KgtkFormat.LANGUAGE_QUALIFIED_STRING_SIGIL:
                                if (lf == "ANY" or label_value.endswith(lf)):
                                    process_label = True
                                    break

                            else:
                                process_label = False
                                break # Do not allow values that are neither strings nor lqstrings.
                        else:
                            process_label = False


                        if process_label:
                            label_priorities[label_key] = priority
                            if label_key in labels:
                                del labels[label_key] # Remove the prior label
                    else:
                        sigil = label_value[0]
                        if sigil == KgtkFormat.STRING_SIGIL:
                            process_label = allow_all_strings

                        elif sigil == KgtkFormat.LANGUAGE_QUALIFIED_STRING_SIGIL:
                            process_label = allow_all_lqstrings or label_value.endswith(language_filter_tuple)

                        else:
                            # Do not allow values that are neither strings nor lqstrings.
                            process_label = False
                else:
                    process_label = True

                if process_label:
                    if label_key in labels:
                        # This label already exists in the table.
                        if self.suppress_duplicate_labels:
                            # Build a list eliminating duplicate elements.
                            # print("Merge '%s' and '%s'" % (key_value, labels[key]), file=self.error_file, flush=True)
                            labels[label_key] = KgtkValue.merge_values(labels[label_key], label_value)
                        else:
                            labels[label_key] = KgtkFormat.LIST_SEPARATOR.join((labels[label_key], label_value))
                        if label_rows is not None:
                            label_rows[label_key].append(row.copy())
                    else:
                        # This is the first instance of this label definition.
                        labels[label_key] = label_value
                        if label_rows is not None:
                            label_rows[label_key] = [ row.copy() ]
                if save_input and not self.remove_label_records:
                    input_rows.append(row.copy())
            else:
                if save_input:
                    input_rows.append(row.copy())
        return labels, input_rows
                
    def load_input_keeping_label_records(self,
                                         kr: KgtkReader,
                                         path: Path,
    )-> typing.List[typing.List[str]]:
        input_rows: typing.List[typing.List[str]] = [ ]

        if self.verbose:
            print("Loading input rows with labels from %s" % path, file=self.error_file, flush=True)
        row: typing.List[str]
        for row in kr:
            input_rows.append(row)
        return input_rows

    def load_input_removing_label_records(self,
                                          kr: KgtkReader,
                                          path: Path,
    )-> typing.List[typing.List[str]]:
        label_select_column_idx: int = self.lookup_label_select_column_idx(kr)
        if label_select_column_idx < 0:
            return self.load_input_keeping_label_records(kr, path)

        if self.verbose:
            print("Loading input rows without labels from %s" % path, file=self.error_file, flush=True)

        input_rows: typing.List[typing.List[str]] = [ ]
        row: typing.List[str]
        for row in kr:
            if row[label_select_column_idx] != self.label_select_column_value:
                input_rows.append(row)

        return input_rows

    def load_input(self,
                   kr: KgtkReader,
                   path: Path,
                   seperate_label_file: bool = False
    )-> typing.List[typing.List[str]]:
        if seperate_label_file:
            return self.load_input_keeping_label_records(kr, path)
        elif self.remove_label_records:
            return self.load_input_removing_label_records(kr, path)
        else:
            return self.load_input_keeping_label_records(kr, path)

    def build_lifted_column_idxs(self,
                                 kr: KgtkReader,
                                 lift_column_idxs: typing.List[int],
                                 input_rows: typing.List[typing.List[str]],
                                 labels: typing.Mapping[str, str],
                                 label_select_column_idx: int,
                                 input_select_column_idx: int,
    )->typing.List[int]:
        """
        Build the lifted column indexes, suppressing those columns
        for which there are are no label values.  This requires a
        pass through the input data.
        """
        if self.verbose:
            print("Checking for empty columns", file=self.error_file, flush=True)
        lift_column_idxs_empties: typing.List[int] = lift_column_idxs.copy()
        lift_column_idx: int
        # Scan the input file, checking for empty output columns.
        for row in input_rows:
            if label_select_column_idx >= 0:
                if row[label_select_column_idx] == self.label_select_column_value:
                    # Skip label records if they have been saved.
                    continue

            if input_select_column_idx >= 0:
                if self.input_select_column_value is not None and row[input_select_column_idx] != self.input_select_column_value:
                    # Not selected for lifting into.
                    continue
            idx: int
            restart: bool = True
            while restart:
                # The restart mechanism compensates for modifying
                # lift_column_idxs_empties inside the for loop, at the
                # expense of potentially double testing some items.
                restart = False
                for idx, lift_column_idx in enumerate(lift_column_idxs_empties):
                    item: str = row[lift_column_idx]
                    if item in labels:
                        lift_column_idxs_empties.pop(idx)
                        restart = True
                        break
            if len(lift_column_idxs_empties) == 0:
                break

        if self.verbose:
            if len(lift_column_idxs_empties) == 0:
                print("No lifted columns are empty", file=self.error_file, flush=True)
            else:
                input_lifting_column_names_empties: typing.List[str] = [ ]
                for idx in lift_column_idxs_empties:
                    input_lifting_column_names_empties.append(kr.column_names[idx])
                print("Unlifted columns: %s" % " ".join(input_lifting_column_names_empties), file=self.error_file, flush=True)

        lifted_column_idxs: typing.List[int] = [ ]
        for lift_column_idx in lift_column_idxs:
            if lift_column_idx not in lift_column_idxs_empties:
                lifted_column_idxs.append(lift_column_idx)            
        return lifted_column_idxs

    def write_output_row(self,
                         ew: KgtkWriter,
                         row: typing.List[str],
                         new_columns: int,
                         input_select_column_idx: int,
                         label_select_column_idx: int,
                         labels: typing.Mapping[str, str],
                         lifted_column_idxs: typing.List[int],
                         lifted_output_column_idxs: typing.List[int],
                         urkw: typing.Optional[KgtkWriter],
                         mlkw: typing.Optional[KgtkWriter],
                         label_rows: typing.Optional[typing.Mapping[str, typing.List[typing.List[str]]]],
                         matched_labels: typing.Set[str],
        )->typing.Tuple[bool, bool, bool]:
        output_row: typing.List[str] = row.copy()
        if new_columns > 0:
            output_row.extend([self.default_value] * new_columns)
        output_select_column_idx: int = input_select_column_idx
                
        do_write: bool = True
        do_lift: bool = True
        did_lift: bool = False
        modified: bool = False
        skipped: bool = False
        if label_select_column_idx >= 0:
            print("label_select_column_idx %d" % label_select_column_idx)
            if row[label_select_column_idx]  == self.label_select_column_value:
                # Don't lift label columns, if we have stored labels in the input records.
                do_lift = False
                if self.remove_label_records:
                    do_write = False
        if input_select_column_idx >= 0:
            if self.input_select_column_value is not None and row[input_select_column_idx] != self.input_select_column_value:
                # Not selected for lifting into.
                do_lift = False
                skipped = True
        if do_lift:
            # Lift the specified columns in this row.
            lifted_column_idx: int
            for idx, lifted_column_idx in enumerate(lifted_column_idxs):
                lifted_output_column_idx: int = lifted_output_column_idxs[idx]
                if self.clear_before_lift:
                    output_row[lifted_output_column_idx] = self.default_value
                if self.overwrite or output_row[lifted_output_column_idxs[idx]] == self.default_value:
                    label_key: str = row[lifted_column_idx]
                    if label_key in labels:
                        label_value: str = labels[label_key]
                        if label_value != output_row[lifted_output_column_idx]:
                            modified = True
                        output_row[lifted_output_column_idx] = label_value
                        did_lift = True # What if we want to note if we lifted all columns?

                        if label_key not in matched_labels:
                            matched_labels.add(label_key)
                            if label_rows is not None and mlkw is not None:
                                label_row: typing.List[str]
                                for label_row in label_rows[label_key]:
                                    mlkw.write(label_row)

            if did_lift and output_select_column_idx >= 0 and self.output_select_column_value is not None:
                if output_row[output_select_column_idx] != self.output_select_column_value:
                    modified = True
                output_row[output_select_column_idx] = self.output_select_column_value

        if do_write:
            if self.output_only_modified_rows:
                if modified:
                    ew.write(output_row)
            else:
                ew.write(output_row)

        if urkw is not None:
            # The unmodified row output file gets all unmodified input rows,
            # including input rows that are labels or are not selected for lift.
            if not modified:
                urkw.write(row)

        return do_write, modified, skipped

    def build_output_column_names(self, ikr: KgtkReader, lifted_column_idxs: typing.List[int])->typing.Tuple[typing.List[str], typing.List[int]]:
        # Build the output column names.
        output_column_names: typing.List[str] = ikr.column_names.copy()
        lifted_output_column_idxs: typing.List[int] = [ ]
        lifted_idx: int
        column_idx: int
        for lifted_idx, column_idx in enumerate(lifted_column_idxs):
            lifted_column_name: str
            if self.output_lifted_column_names is not None and lifted_idx < len(self.output_lifted_column_names):
                lifted_column_name = self.output_lifted_column_names[lifted_idx]
            else:
                lifted_column_name = ikr.column_names[column_idx] + self.output_lifted_column_suffix

            if lifted_column_name in ikr.column_name_map:
                # Overwrite an existing lifted output column.
                #
                # TODO: DO we want to control whether or not we overwrite existing columns?
                lifted_output_column_idxs.append(ikr.column_name_map[lifted_column_name])
            else:
                # Append a new lifted output column.
                lifted_output_column_idxs.append(len(output_column_names))
                output_column_names.append(lifted_column_name)
        return output_column_names, lifted_output_column_idxs

    def open_output_writer(self,
                           ikr: KgtkReader,
                           lifted_column_idxs: typing.List[int]
    )->typing.Tuple[KgtkWriter, typing.List[int]]:
        # Build the output column names.
        output_column_names: typing.List[str]
        lifted_output_column_idxs: typing.List[int]
        output_column_names, lifted_output_column_idxs = self.build_output_column_names(ikr, lifted_column_idxs)

        if self.verbose:
            print("Opening the output file: %s" % self.output_file_path, file=self.error_file, flush=True)
        ew: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode[ikr.mode.name],
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         use_mgzip=False if self.input_reader_options is None else self.input_reader_options.use_mgzip , # Hack!
                                         mgzip_threads=3 if self.input_reader_options is None else self.input_reader_options.mgzip_threads , # Hack!
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)        

        return ew, lifted_output_column_idxs

    def build_labels_needed(self,
                            input_rows: typing.List[typing.List[str]],
                            input_select_column_idx: int,
                            lift_column_idxs: typing.List[int],
    )->typing.Set[str]:
        labels_needed: typing.Set[str] = set()

        row: typing.List[str]
        for row in input_rows:
            if input_select_column_idx >= 0:
                if self.input_select_column_value is not None and row[input_select_column_idx] != self.input_select_column_value:
                    # Not selected for lifting into.
                    continue
            lift_column_idx: int
            for lift_column_idx in lift_column_idxs:
                label_key: str = row[lift_column_idx]
                labels_needed.add(label_key)

        if self.verbose:
            print("Labels needed: %d" % len(labels_needed), file=self.error_file, flush=True)

        return labels_needed

    def process_in_memory(self, ikr: KgtkReader,
                          lkr: typing.Optional[KgtkReader],
                          urkw: typing.Optional[KgtkWriter],
                          mlkw: typing.Optional[KgtkWriter],
                          ulkw: typing.Optional[KgtkWriter],
                          ):
        """
        Process the lift using in-memory buffering.  The labels will added to a
        dict in memory, and depending upon the options selected, the input
        rows may be kept on a list in memory.

        """
        if self.verbose:
            print("Lifting with in-memory buffering.", file=self.error_file, flush=True)
        lift_column_idxs: typing.List[int] = self.build_lift_column_idxs(ikr)

        labels: typing.Mapping[str, str] = { }
        input_rows: typing.Optional[typing.List[typing.List[str]]] = None

        label_rows: typing.Optional[typing.MutableMapping[str, typing.List[typing.List[str]]]]

        if mlkw is None and ulkw is None:
            label_rows = None
        else:
            label_rows = { }
        
        input_select_column_idx: int
        if self.input_select_column_value is not None or self.output_select_column_value is not None:
            input_select_column_idx = self.lookup_input_select_column_idx(ikr)
        else:
            input_select_column_idx = -1

        # Unless told otherwise, assume that label rows won't be saved
        # in the input rows:
        label_select_column_idx: int = -1

        # Extract the labels, and maybe store the input rows.
        if lkr is not None and self.label_file_path is not None:
            labels_needed: typing.Optional[typing.Set[str]] = None
            if self.prefilter_labels:
                if self.verbose:
                    print("Reading input data to prefilter the labels.", file=self.error_file, flush=True)
                input_rows = self.load_input(ikr, self.input_file_path, seperate_label_file=True)
                labels_needed = self.build_labels_needed(input_rows, input_select_column_idx, lift_column_idxs)
            # Read the label file.
            if self.verbose:
                print("Loading labels from the label file.", file=self.error_file, flush=True)
            # We don't need to worry about input rows in the label file.
            labels, _ = self.load_labels(lkr, self.label_file_path, save_input=False, labels_needed=labels_needed, label_rows=label_rows, is_label_file=True)
        else:
            if self.verbose:
                print("Loading labels and reading data from the input file.", file=self.error_file, flush=True)
            # Read the input file, extracting the labels. The label
            # records may or may not be saved in the input rows, depending
            # upon whether we plan to pass them through to the output.
            labels, input_rows = self.load_labels(ikr, self.input_file_path, label_rows=label_rows)

            if not self.remove_label_records:
                # Save the label column index in the input rows:
                label_select_column_idx = self.lookup_label_select_column_idx(ikr)

        label_count: int = len(labels)
        if label_count == 0 and not self.ok_if_no_labels:
            raise ValueError("No labels were found.")

        lifted_column_idxs: typing.List[int]
        if self.suppress_empty_columns:
            if input_rows is None:
                # We need to read the input records now in order to determine
                # which lifted columns must be suppressed.
                if self.verbose:
                    print("Reading input data to suppress empty columns.", file=self.error_file, flush=True)
                input_rows = self.load_input(ikr, self.input_file_path)
            lifted_column_idxs = self.build_lifted_column_idxs(ikr, lift_column_idxs, input_rows, labels, label_select_column_idx, input_select_column_idx)
        else:
            # Lift all the candidate columns.
            lifted_column_idxs = lift_column_idxs.copy()

        ew: KgtkWriter
        lifted_output_column_idxs: typing.List[int]
        ew, lifted_output_column_idxs = self.open_output_writer(ikr, lifted_column_idxs)

        if self.verbose:
            print("Writing output records", file=self.error_file, flush=True)

        matched_labels: typing.Set[str] = set()
        new_columns: int = len(ew.column_names) - len(ikr.column_names)
        input_line_count: int = 0
        input_skipped_count: int = 0
        output_line_count: int = 0
        output_modified_count: int = 0
        if input_rows is None:
            do_write: bool
            modified: bool
            skipped: bool
            # Read the input file and process it in one pass:
            for row in ikr:
                input_line_count += 1
                do_write, modified, skipped = self.write_output_row(ew,
                                                                    row,
                                                                    new_columns,
                                                                    input_select_column_idx,
                                                                    label_select_column_idx,
                                                                    labels,
                                                                    lifted_column_idxs,
                                                                    lifted_output_column_idxs,
                                                                    urkw,
                                                                    mlkw,
                                                                    label_rows,
                                                                    matched_labels,
                                                                    )
                if do_write:
                    output_line_count += 1
                    if modified:
                        output_modified_count += 1
                if skipped:
                    input_skipped_count += 1
        else:
            # Use the stored input records:
            for row in input_rows:
                input_line_count += 1
                do_write, modified, skipped = self.write_output_row(ew,
                                                                    row,
                                                                    new_columns,
                                                                    input_select_column_idx,
                                                                    label_select_column_idx,
                                                                    labels,
                                                                    lifted_column_idxs,
                                                                    lifted_output_column_idxs,
                                                                    urkw,
                                                                    mlkw,
                                                                    label_rows,
                                                                    matched_labels,
                                                                    )
                if do_write:
                    output_line_count += 1
                    if modified:
                        output_modified_count += 1
                if skipped:
                    input_skipped_count += 1

        if ulkw is not None and label_rows is not None:
            label_key: str
            for label_key in sorted(label_rows.keys()):
                if label_key not in matched_labels:
                    label_row: typing.List[str]
                    for label_row in label_rows[label_key]:
                        ulkw.write(label_row)

        if self.verbose:
            print("Read %d non-label input records, %d skipped." % (input_line_count, input_skipped_count), file=self.error_file, flush=True)
            print("%d labels were found, %d matched." % (label_count, len(matched_labels)), file=self.error_file, flush=True)
            print("Wrote %d output records, %d modified." % (output_line_count, output_modified_count), file=self.error_file, flush=True)
        
        ew.close()
    
    def process_as_merge(self, ikr: KgtkReader,
                         lkr: KgtkReader,
                         urkw: typing.Optional[KgtkWriter],
                         mlkw: typing.Optional[KgtkWriter],
                         ulkw: typing.Optional[KgtkWriter],
                         ):
        """
        Process the lift as a merge between two sorted files.

        """
        if self.verbose:
            print("Merging sorted input and label files.", file=self.error_file, flush=True)
        lift_column_idxs: typing.List[int] = self.build_lift_column_idxs(ikr)
        if len(lift_column_idxs) != 1:
            raise ValueError("Expecting exactly one lift_column_idxs, got %d" % len(lift_column_idxs))

        # Setup for language filtering:
        language_filtering: bool = self.languages is not None and len(self.languages) > 0
        allow_all_strings: bool = False
        allow_all_lqstrings: bool = False
        language_filter: typing.List[str] = list()
        language_filter_tuple: typing.Tuple[str, ...] = ()
        if language_filtering:
            for lf in self.languages:
                if lf == 'NONE':
                    if self.prioritize:
                        language_filter.append(lf)
                    else:
                        allow_all_strings = True
                elif lf == 'ANY':
                    if self.prioritize:
                        language_filter.append(lf)
                    else:
                        allow_all_lqstrings = True
                else:
                    language_filter.append("'@" + lf)
            language_filter_tuple = tuple(language_filter)

        ew: KgtkWriter
        lifted_output_column_idxs: typing.List[int]
        ew, lifted_output_column_idxs = self.open_output_writer(ikr, lift_column_idxs)

        new_columns: int = len(ew.column_names) - len(ikr.column_names)
        if new_columns not in (0, 1):
            raise ValueError("Expecting zero or one new columns, got %d" % new_columns)

        lift_column_idx: int = lift_column_idxs[0] # For convenience
        lifted_output_column_idx: int = lifted_output_column_idxs[0] # For convenience

        label_match_column_idx: int
        label_select_column_idx: int
        label_value_column_idx: int
        label_match_column_idx, label_select_column_idx, label_value_column_idx = self.lookup_label_table_idxs(lkr)

        input_select_column_idx: int = -1
        if self.input_select_column_value is not None or self.output_select_column_value is not None:
            input_select_column_idx = self.lookup_input_select_column_idx(ikr)
            
        input_line_count: int = 0
        input_skipped_count: int = 0
        label_line_count: int = 0
        label_match_count: int = 0
        output_line_count: int = 0
        output_modified_count: int = 0

        current_label_row: typing.Optional[typing.List[str]] = None
        more_labels: bool = True
        # Read the first label record.
        try:
            current_label_row = lkr.nextrow()
            label_line_count += 1
        except StopIteration:
            more_labels = False

        # We carry last_value_to_lift and lifted_label_value over
        # iterations in case the input file has multiple records with
        # the same value to lift.
        last_value_to_lift: typing.Optional[str] = None
        lifted_label_value: str = self.default_value

        if self.verbose:
            print("Processing the input records.", file=self.error_file, flush=True)

        output_row: typing.List[str]
        row: typing.List[str]
        for row in ikr:
            input_line_count += 1

            if self.input_select_column_value is not None and input_select_column_idx >= 0:
                # Skip input rows that do not meet the selection criterion.
                if row[input_select_column_idx] != self.input_select_column_value:
                    input_skipped_count += 1
                    if not self.output_only_modified_rows:
                        if new_columns > 0:
                            output_row = row.copy()
                            output_row.append(self.default_value)
                        else:
                            output_row = row
                        ew.write(output_row)
                    continue

            value_to_lift: str = row[lift_column_idx]
            if last_value_to_lift is None or value_to_lift != last_value_to_lift:
                last_value_to_lift = value_to_lift
                lifted_label_value = self.default_value

                # Read label records until we come to the first record that
                # has a node1 value equal to or greater than the value we we want to lift.
                while more_labels and current_label_row is not None and current_label_row[label_match_column_idx] < value_to_lift:
                    if ulkw is not None:
                        ulkw.write(current_label_row)
                    try:
                        current_label_row = lkr.nextrow()
                        label_line_count += 1
                    except StopIteration:
                        more_labels = False
                        break

                # While the label records have node1 values equal to the value we are trying to lift,
                # look for label values from the label file.
                current_priority: int = 9999999 # Large numbers are low priority.
                while more_labels and current_label_row is not None and current_label_row[label_match_column_idx] == value_to_lift:
                    if label_select_column_idx < 0 or current_label_row[label_select_column_idx] == self.label_select_column_value:
                        label_value: str = current_label_row[label_value_column_idx]

                        process_label: bool
                        sigil: str
                        if language_filtering:
                            if len(label_value) == 0:
                                # Empty values are not valid for language filtering.
                                process_label = False
                                
                            elif self.prioritize:
                                # We will not use a label value unless it passes one of
                                # the priority-ordered filters.
                                process_label = False
                                sigil = label_value[0]
                                priority: int
                                lf: str
                                for priority, lf in enumerate(language_filter):
                                    if priority >= current_priority:
                                        # A priori label is higher priority.
                                        break

                                    elif sigil == KgtkFormat.STRING_SIGIL:
                                        # This is a KGTK string without language qualification.
                                        if lf == "NONE":
                                            # We will allow it.
                                            current_priority = priority
                                            process_label = True
                                            break

                                    elif sigil == KgtkFormat.LANGUAGE_QUALIFIED_STRING_SIGIL:
                                        # This is a KGTK language-qualified string.
                                        if lf == "ANY" or label_value.endswith(lf):
                                            # We will allow it.
                                            current_priority = priority
                                            process_label = True
                                            break

                                    else:
                                        # Do not allow values that are neither strings nor lqstrings.
                                        break
                                            
                            else:
                                # Optimized code for non-prioritized langiage selection.
                                sigil = label_value[0]
                                if sigil == KgtkFormat.STRING_SIGIL:
                                    process_label = allow_all_strings

                                elif sigil == KgtkFormat.LANGUAGE_QUALIFIED_STRING_SIGIL:
                                    process_label = allow_all_lqstrings or label_value.endswith(language_filter_tuple)

                                else:
                                    # Do not allow values that are neither strings nor lqstrings.
                                    process_label = False
                        else:
                            process_label = True

                        if process_label:
                            # TODO: is this code positioned correctly?
                            if mlkw is not None:
                                mlkw.write(current_label_row)
                            label_match_count += 1
                                
                            if lifted_label_value != self.default_value:
                                if self.suppress_duplicate_labels:
                                    lifted_label_value = KgtkValue.merge_values(lifted_label_value, label_value)
                                else:
                                    lifted_label_value = KgtkFormat.LIST_SEPARATOR.join((lifted_label_value, label_value))
                            else:
                                lifted_label_value = label_value

                    try:
                        current_label_row = lkr.nextrow()
                        label_line_count += 1
                    except StopIteration:
                        more_labels = False
                        break

            output_row = row.copy()
            modified: bool = False
            if new_columns > 0:
                output_row.append(self.default_value)
            if self.clear_before_lift:
                output_row[lifted_output_column_idx] = self.default_value
            if self.overwrite or output_row[lifted_output_column_idx] != self.default_value:
                if lifted_label_value != self.default_value:
                    if output_row[lifted_output_column_idx] != lifted_label_value:
                        modified = True
                    output_row[lifted_output_column_idx] = lifted_label_value
                    if self.output_select_column_value is not None and input_select_column_idx >= 0:
                        if output_row[input_select_column_idx] != self.output_select_column_value:
                            modified = True
                        output_row[input_select_column_idx] = self.output_select_column_value

            if self.output_only_modified_rows:
                if modified:
                    ew.write(output_row)
                    output_line_count += 1
                    output_modified_count += 1
            else:
                ew.write(output_row)
                output_line_count += 1
                if modified:
                    output_modified_count += 1

            if urkw is not None:
                if not modified:
                    urkw.write(row)

        while more_labels:
            if current_label_row is not None and ulkw is not None:
                ulkw.write(current_label_row)
            try:
                current_label_row = lkr.nextrow()
                label_line_count += 1
            except StopIteration:
                more_labels = False
                break
        lkr.close()

        if self.verbose:
            print("Read %d input records, %d skipped." % (input_line_count, input_skipped_count), file=self.error_file, flush=True)
            print("Read %d label records, %d matched." % (label_line_count, label_match_count), file=self.error_file, flush=True)
            print("Wrote %d output records, %d modified." % (output_line_count, output_modified_count), file=self.error_file, flush=True)
             
        ew.close()

    
    def process(self):
        # Open the input file.
        input_mode: typing.Optional[KgtkReaderMode] = KgtkReaderMode.NONE if self.force_input_mode_none else None
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        ikr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                           error_file=self.error_file,
                                           mode=input_mode,
                                           options=self.input_reader_options,
                                           value_options = self.value_options,
                                           verbose=self.verbose,
                                           very_verbose=self.very_verbose,
        )

        lkr: typing.Optional[KgtkReader] = None
        if self.label_file_path is not None:
            if self.verbose:
                if self.input_file_path is not None:
                    print("Opening the label file: %s" % self.label_file_path, file=self.error_file, flush=True)
                else:
                    print("Reading the label data from stdin", file=self.error_file, flush=True)

            lkr =  KgtkReader.open(self.label_file_path,
                                   error_file=self.error_file,
                                   options=self.label_reader_options,
                                   value_options = self.value_options,
                                   verbose=self.verbose,
                                   very_verbose=self.very_verbose,
            )

        urkw: typing.Optional[KgtkWriter] = None
        if self.unmodified_row_file_path is not None:
            urkw = KgtkWriter.open(ikr.column_names,
                                   self.unmodified_row_file_path,
                                   mode=KgtkWriter.Mode[ikr.mode.name],
                                   require_all_columns=False,
                                   prohibit_extra_columns=True,
                                   fill_missing_columns=True,
                                   use_mgzip=False if self.input_reader_options is None else self.input_reader_options.use_mgzip , # Hack!
                                   mgzip_threads=3 if self.input_reader_options is None else self.input_reader_options.mgzip_threads , # Hack!
                                   gzip_in_parallel=False,
                                   verbose=self.verbose,
                                   very_verbose=self.very_verbose)        
            
        label_column_names: typing.List[str]
        label_file_mode_name: str

        mlkw: typing.Optional[KgtkWriter] = None
        if self.matched_label_file_path is not None:
            label_column_names: typing.List[str]
            label_file_mode_name: str
            if lkr is not None:
                label_column_names = lkr.column_names
                label_file_mode_name = lkr.mode.name
            else:
                label_column_names = ikr.column_names
                label_file_mode_name = ikr.mode.name
            mlkw = KgtkWriter.open(label_column_names,
                                   self.matched_label_file_path,
                                   mode=KgtkWriter.Mode[label_file_mode_name],
                                   require_all_columns=False,
                                   prohibit_extra_columns=True,
                                   fill_missing_columns=True,
                                   use_mgzip=False if self.input_reader_options is None else self.input_reader_options.use_mgzip , # Hack!
                                   mgzip_threads=3 if self.input_reader_options is None else self.input_reader_options.mgzip_threads , # Hack!
                                   gzip_in_parallel=False,
                                   verbose=self.verbose,
                                   very_verbose=self.very_verbose)        
            

        ulkw: typing.Optional[KgtkWriter] = None
        if self.unmatched_label_file_path is not None:
            label_column_names: typing.List[str]
            label_file_mode_name: str
            if lkr is not None:
                label_column_names = lkr.column_names
                label_file_mode_name = lkr.mode.name
            else:
                label_column_names = ikr.column_names
                label_file_mode_name = ikr.mode.name
            ulkw = KgtkWriter.open(label_column_names,
                                   self.unmatched_label_file_path,
                                   mode=KgtkWriter.Mode[label_file_mode_name],
                                   require_all_columns=False,
                                   prohibit_extra_columns=True,
                                   fill_missing_columns=True,
                                   use_mgzip=False if self.input_reader_options is None else self.input_reader_options.use_mgzip , # Hack!
                                   mgzip_threads=3 if self.input_reader_options is None else self.input_reader_options.mgzip_threads , # Hack!
                                   gzip_in_parallel=False,
                                   verbose=self.verbose,
                                   very_verbose=self.very_verbose)        
            
        if self.input_lifting_column_names is not None and len(self.input_lifting_column_names) == 1 and \
           not self.suppress_empty_columns and \
           self.input_is_presorted and \
           self.labels_are_presorted and \
           lkr is not None:
            self.process_as_merge(ikr, lkr, urkw, mlkw, ulkw)
        else:
            self.process_in_memory(ikr, lkr, urkw, mlkw, ulkw)

        if urkw is not None:
            urkw.close()

        if mlkw is not None:
            mlkw.close()

        if ulkw is not None:
            ulkw.close()

def main():
    """
    Test the KGTK lift processor.

    TODO: the unmodified row file path.
    TODO: there are other missing options.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data", type=Path, default="-")

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--label-file", dest="label_file_path",
                              help="An optional KGTK file with label records (default=%(default)s).", type=Path, default=None)


    parser.add_argument(      "--input-select-column", "--input-label-column", dest="input_select_column_name",
                              help="If input record selection is enabled by --input-select-value, " +
                              "the name of a column that determines which records received lifted values. " +
                              "The default is the 'label' column or its alias.", default=None)

    parser.add_argument(      "--input-select-value", "--input-label-value", "--target-label-value", dest="input_select_column_value",
                              help="The value in the input select column that identifies a record to receive lifted values. " +
                              "The default is not to perform input record selection, " +
                              "and all input records except label records may receive lifted values. ",
                              default=None)
    
    parser.add_argument(      "--columns-to-lift", dest="input_lifting_column_names",
                              help="The columns for which matching labels are to be lifted. " +
                              "The default is [node1, label, node2] or their aliases.", nargs='*')

    parser.add_argument(      "--columns-to-write", dest="output_lifted_column_names",
                              help="The columns into which to store the lifted values. " +
                              "The default is [node1;label, label;label, node2;label] or their aliases.", nargs='*')

    parser.add_argument(      "--lift-suffix", dest="output_lifted_column_suffix",
                              help="The suffix used for newly created output columns. (default=%(default)s).",
                              default=KgtkLift.DEFAULT_OUTPUT_LIFTED_COLUMN_SUFFIX)

    parser.add_argument(      "--default-value", dest="default_value",
                              help="The value to use if a lifted label is not found. (default=%(default)s)", default="")

    parser.add_argument(      "--update-select-value", "--target-new-label-value", dest="output_select_column_value",
                              help="A new value for the select (label) column for records that received lifted values. " +
                              "The default is not to update the select(label) column.", default=None)
    

    parser.add_argument(      "--label-select-column", "--label-name", dest="label_select_column_name",
                              help="The name of the column that contains a special value that identifies label records. " +
                              "The default is 'label' or its alias.", default=None)

    parser.add_argument("-p", "--label-select-value", "--label-value", "--property", dest="label_select_column_value",
                              help="The special value in the label select column that identifies a label record. " +
                              "(default=%(default)s).", default=KgtkLift.DEFAULT_LABEL_SELECT_COLUMN_VALUE)
    
    parser.add_argument(      "--label-match-column", "--node1-name", dest="label_match_column_name",
                              help="The name of the column in the label records that contains the value " +
                              "that matches the value in a column being lifted in the input records. " +
                              "The default is 'node1' or its alias.", default=None)

    parser.add_argument(      "--label-value-column", "--node2-name", "--lift-from", dest="label_value_column_name",
                              help="The name of the column in the label record that contains the value " +
                              "to be lifted into the input record that is receiving lifted values. " +
                              "The default is 'node2' or its alias.", default=None)

    parser.add_argument(      "--remove-label-records", dest="remove_label_records",
                              help="If true, remove label records from the output. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--sort-lifted-labels", dest="sort_lifted_labels",
                              help="If true, sort lifted labels with lists. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-duplicate-labels", dest="suppress_duplicate_labels",
                              help="If true, suppress duplicate values in lifted labels with lists (implies sorting). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-empty-columns", dest="suppress_empty_columns",
                              help="If true, do not create new columns that would be empty. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--ok-if-no-labels", dest="ok_if_no_labels",
                              help="If true, do not abort if no labels were found. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--prefilter-labels", dest="prefilter_labels",
                              help="If true, read the input file before reading the label file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--input-file-is-presorted", dest="input_is_presorted",
                              help="If true, the input file is presorted on the column for which values are to be lifted. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--label-file-is-presorted", dest="labels_are_presorted",
                              help="If true, the label file is presorted on the node1 column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--clear-before-lift", dest="clear_before_lift",
                              help="If true, set columns to write to the default value before lifting. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--overwrite", dest="overwrite",
                              help="If true, overwrite non-default values in the columns to write. " +
                              "If false, do not overwrite non-default values in the columns to write. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--output-only-modified-rows", dest="output_only_modified_rows",
                              help="If true, output only modified edges to the primary output stream. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    KgtkReader.add_debug_arguments(parser)
    # TODO: seperate reader options for the label file.
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="label", defaults=False)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="input", fallback=True)
    label_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="label", fallback=True)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

    # Show the final option structures for debugging and documentation.   
    if args.show_options:
        print("input: %s" % str(args.input_file_path), file=error_file, flush=True)
        if args.label_file_path is not None:
            print("--label-file=%s" % str(args.label_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)

        if args.input_select_column_name is not None:
            print("--input-select-column=%s" % args.input_select_column_name, file=error_file, flush=True)
        if args.input_select_column_value is not None:
            print("--input-select-value=%s" % args.input_select_column_value, file=error_file, flush=True)
        if args.input_lifting_column_names is not None and len(args.input_lifting_column_names) > 0:
            print("--columns-to-lift %s" % " ".join(args.input_lifting_column_names), file=error_file, flush=True)
        if args.output_lifted_column_names is not None and len(args.output_lifted_column_names) > 0:
            print("--columns-to-write %s" % " ".join(args.output_lifted_column_names), file=error_file, flush=True)

        print("--lift-suffix=%s" % args.output_lifted_column_suffix, file=error_file, flush=True)
        if args.output_select_column_value is not None:
            print("--update-label-value=%s" % args.output_select_column_value, file=error_file, flush=True)

        if args.label_select_column_name is not None:
            print("--label-select-column=%s" % args.label_select_column_name, file=error_file, flush=True)
        if args.label_select_column_value is not None:
            print("--label-select-value=%s" % args.label_select_column_value, file=error_file, flush=True)
        if args.label_match_column_name is not None:
            print("--label-match-column=%s" % args.label_match_column_name, file=error_file, flush=True)
        if args.label_value_column_name is not None:
            print("--label-value-column=%s" % args.label_value_column_name, file=error_file, flush=True)


        print("--default-value=%s" % repr(args.default_value))
        print("--remove-label-records=%s" % repr(args.remove_label_records))
        print("--sort-lifted-labels-labels=%s" % repr(args.sort_lifted_labels))
        print("--suppress-duplicate-labels=%s" % repr(args.suppress_duplicate_labels))
        print("--suppress-empty-columns=%s" % repr(args.suppress_empty_columns))
        print("--ok-if-no-labels=%s" % repr(args.ok_if_no_labels))
        print("--prefilter-labels=%s" % repr(args.prefilter_labels))
        print("--input-file-is-presorted=%s" % repr(args.input_is_presorted))
        print("--label-file-is-presorted=%s" % repr(args.labels_are_presorted))
        print("--clear-before-lift=%s" % repr(args.clear_before_lift))
        print("--overwrite=%s" % repr(args.overwrite))
        print("--output-only-modified-rows=%s" % repr(args.output_only_modified_rows))
        
        input_reader_options.show(out=error_file, who="input")
        label_reader_options.show(out=error_file, who="label")
        value_options.show(out=error_file)

    kl: KgtkLift = KgtkLift(
        input_file_path=args.input_file_path,
        label_file_path=args.label_file_path,
        output_file_path=args.output_file_path,

        input_select_column_name=args.input_select_column_name,
        input_select_column_value=args.input_select_column_value, 
        input_lifting_column_names=args.input_lifting_column_names,

        output_select_column_value=args.output_select_column_value, 
        output_lifted_column_suffix=args.output_lifted_column_suffix,
        output_lifted_column_names=args.output_lifted_column_names,

        label_select_column_name=args.label_select_column_name,
        label_select_column_value=args.label_select_column_value,
        label_match_column_name=args.label_match_column_name,
        label_value_column_name=args.label_value_column_name,

        default_value=args.default_value,
        
        remove_label_records=args.remove_label_records,
        sort_lifted_labels=args.sort_lifted_labels,
        suppress_duplicate_labels=args.suppress_duplicate_labels,
        suppress_empty_columns=args.suppress_empty_columns,
        ok_if_no_labels=args.ok_if_no_labels,
        prefilter_labels=args.prefilter_labels,
        input_is_presorted=args.input_is_presorted,
        labels_are_presorted=args.labels_are_presorted,

        clear_before_lift=args.clear_before_lift,
        overwrite=args.overwrite,

        input_reader_options=input_reader_options,
        label_reader_options=label_reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    kl.process()

if __name__ == "__main__":
    main()



"""
Export a KGTK file to Graph-tool format.

Note:  the log file wasn't coverted to the new filename parsingAPI.

Note:  The input file is read twice: once for the header, and once for the
data.  Thus, stdin cannot be used as the input file.

TODO: Convert to KgtkReader and read the file only once.
"""
from argparse import Namespace
import typing
import pandas as pd

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Export a KGTK file to Graph-tool format.'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True, optional=False)
    parser.add_output_file(who="Graph tool file to dump the graph too - if empty, it will not be saved.", optional=True)

    parser.add_argument('--undirected', dest="undirected",
                        help="When True, the graph is undirected. (default=%(default)s)",
                        type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument('--node-file', dest='node_file', type=str,
                        default=None,
                        help="Specify the location of node file.")
    
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        undirected: bool,
        node_file: None,

        errors_to_stdout: bool,
        errors_to_stderr: bool,
        show_options: bool,
        verbose: bool,
        very_verbose: bool,


        **kwargs, # Whatever KgtkFileOptions and KgtkValueOptions want.
        ):
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import KGTKException
    import kgtk.gt.analysis_utils as gtanalysis
    from kgtk.gt.gt_load import load_graph_from_kgtk
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    try:
        # Select where to send error messages, defaulting to stderr.
        error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

        # Build the option structures.
        reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
        value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

        input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
        output_gt_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(output_file)

        if verbose:
            print('loading the KGTK input file...\n', file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         error_file=error_file,
                                         options=reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
                                         )
        sub: int = kr.get_node1_column_index()
        if sub < 0:
            print("Missing node1 (subject) column.", file=error_file, flush=True)
        pred: int = kr.get_label_column_index()
        if pred < 0:
            print("Missing label (predicate) column.", file=error_file, flush=True)
        obj: int = kr.get_node2_column_index()
        if obj < 0:
            print("Missing node2 (object) column", file=error_file, flush=True)
        if sub < 0 or pred < 0 or obj < 0:
            kr.close()
            raise KGTKException("Exiting due to missing columns.")

        G2 = load_graph_from_kgtk(kr, directed=not undirected, ecols=(sub, obj), verbose=verbose, out=error_file)

        if node_file is not None:
            kr_node: KgtkReader = KgtkReader.open(node_file,
                                                  error_file=error_file,
                                                  options=reader_options,
                                                  value_options=value_options,
                                                  verbose=verbose,
                                                  very_verbose=very_verbose,
                                                  )
            d = {}
            for i in range(0, len(list(G2.vp['name']))):
                d[G2.vp['name'][i]] = i

            vprop_dict = {}
            for col in kr_node.column_name_map:
                if col != 'id':
                    vprop_dict[col] = G2.new_vertex_property("string")
            for row in kr_node:
                for col in kr_node.column_name_map:
                    if col != 'id':
                        v = G2.vertex(d[row[kr_node.column_name_map['id']]])
                        vprop_dict[col][v] = row[kr_node.column_name_map[col]]
            for col in kr_node.column_name_map:
                if col != 'id':
                    G2.vertex_properties[col] = vprop_dict[col]


        if verbose:
            print('graph loaded! It has %d nodes and %d edges.' % (G2.num_vertices(), G2.num_edges()), file=error_file, flush=True)
            print('\n###Top relations:', file=error_file, flush=True)
            for rel, freq in gtanalysis.get_topN_relations(G2, pred_property=kr.column_names[pred]):
                print('%s\t%d' % (rel, freq), file=error_file, flush=True)

        if output_gt_file is not None:
            if verbose:
                print('\nNow saving the graph to %s' % str(output_gt_file), file=error_file, flush=True)
            G2.save(str(output_gt_file))
            if verbose:
                print('Done saving the graph.', file=error_file, flush=True)
    except Exception as e:
        raise KGTKException('Error: ' + str(e))



"""Filter a KGTK file based on whether one or more records exist in a second
KGTK file with matching values for one or more fields.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Count unique values in a column.',
        'description': 'Count the unique values in a column in a KGTK file. Write the unique values and counts as a new KGTK file.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert unique --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.join.unique import Unique
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument('-c', "--column", "--columns", dest="column_names", nargs='*',
                        metavar="COLUMN_NAME",
                              help="The column(s) to count unique values (default=node2 or its alias).")

    parser.add_argument(      "--empty", dest="empty_value", help="A value to substitute for empty values (default=%(default)s).", default="")

    parser.add_argument(      "--label", dest="label_value", help="The output file label column value (default=%(default)s).", default="count")

    # TODO: use an emum
    parser.add_argument(      "--format", dest="output_format", help="The output file format and mode (default=%(default)s).",
                              default=Unique.DEFAULT_FORMAT, choices=Unique.OUTPUT_FORMATS)

    parser.add_argument(      "--prefix", dest="prefix", help="The value prefix (default=%(default)s).", default="")

    parser.add_argument(      "--where", dest="where_column_name",
                              help="The name of a column for a record selection test. (default=%(default)s).", default=None)

    parser.add_argument(      "--in", dest="where_values", nargs="+",
                              help="The list of values for a record selection test. (default=%(default)s).", default=None)

    parser.add_argument(      "--value-filter", dest="value_filter", metavar="VALUE_FILTER_RE",
                              help="A regular expression filter on the extracted values. (default=%(default)s).", default="")

    parser.add_argument(      "--value-match-type", dest="value_match_type", 
                              help="Which type of regular expression value match: %(choices)s. (default=%(default)s).",
                              type=str, default="match", choices=["fullmatch", "match", "search"])

    parser.add_argument(      "--presorted", dest="presorted", metavar="True|False",
                              help="When True, the input file is presorted. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--min-count", dest="min_count", type=int, default=0,
                              help=f"The minimum count output filter (does not apply to {repr(Unique.NODE_FORMAT)} format). The minimum and maximum count filters are conjunctive. (default=%(default)s).")

    parser.add_argument(      "--max-count", dest="max_count", type=int, default=999999999999999,
                              help=f"The minimum count output filter (does not apply to {repr(Unique.NODE_FORMAT)} format). The minimum and maximum count filters are conjunctive. (default=%(default)s).")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        column_names: typing.Optional[typing.List[str]] = None,
        empty_value: str = "",
        label_value: str = "count",

        output_format: str = "edge",
        prefix: str = "",

        where_column_name: typing.Optional[str] = None,
        where_values: typing.Optional[typing.List[str]] = None,

        value_filter: str = "",
        value_match_type: str = "match",

        presorted: bool = False,

        min_count: int = 0,
        max_count: int = sys.maxsize,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path

    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.join.unique import Unique
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        if column_names is not None:
            print("--columns %s" % " ".join(column_names), file=error_file, flush=True)
        print("--empty=%s" % str(empty_value), file=error_file)
        print("--label=%s" % str(label_value), file=error_file)
        print("--format=%s" % output_format, file=error_file)
        print("--prefix=%s" % prefix, file=error_file)
        if where_column_name is not None:
            print("--where=%s" % where_column_name, file=error_file)
        if where_values is not None and len(where_values) > 0:
            print("--in=%s" % " ".join(where_values), file=error_file)
        print("--value-filter=%s" % repr(value_filter), file=error_file)
        print("--value-match-type=%s" % repr(value_match_type), file=error_file)
        print("--prefix=%s" % repr(presorted), file=error_file)
        print("--min-count=%d" % min_count, file=error_file)
        print("--max-count=%d" % max_count, file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if min_count > max_count:
        raise KGTKException("Error: --min-count %d is greater than --max-count %d" % (min_count, max_count))

    try:
        uniq: Unique = Unique(
            input_file_path=input_kgtk_file,
            output_file_path=output_kgtk_file,
            column_names=column_names,
            label_value=label_value,
            empty_value=empty_value,
            output_format=output_format,
            prefix=prefix,
            where_column_name=where_column_name,
            where_values=where_values,
            value_filter=value_filter,
            value_match_type=value_match_type,
            presorted=presorted,
            min_count=min_count,
            max_count=max_count,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        uniq.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




wiki_namespaces = {
    'wikibase': 'http://wikiba.se/ontology#',
    'wd': 'http://www.wikidata.org/entity/',
    'wdt': 'http://www.wikidata.org/prop/direct/',
    'wdtn': 'http://www.wikidata.org/prop/direct-normalized/',
    'wdno': 'http://www.wikidata.org/prop/novalue/',
    'wds': 'http://www.wikidata.org/entity/statement/',
    'wdv': 'http://www.wikidata.org/value/',
    'wdref': 'http://www.wikidata.org/reference/',
    'p': 'http://www.wikidata.org/prop/',
    'pr': 'http://www.wikidata.org/prop/reference/',
    'prv': 'http://www.wikidata.org/prop/reference/value/',
    'prn': 'http://www.wikidata.org/prop/reference/value-normalized/',
    'ps': 'http://www.wikidata.org/prop/statement/',
    'psv': 'http://www.wikidata.org/prop/statement/value/',
    'psn': 'http://www.wikidata.org/prop/statement/value-normalized/',
    'pq': 'http://www.wikidata.org/prop/qualifier/',
    'pqv': 'http://www.wikidata.org/prop/qualifier/value/',
    'pqn': 'http://www.wikidata.org/prop/qualifier/value-normalized/',
    'prov': 'http://www.w3.org/ns/prov#',
    'skos': 'http://www.w3.org/2004/02/skos/core#',
    'schema': 'http://schema.org/'
}



"""
Generate wikidata triples from two a kgtk edge file

"""
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles


def parser():
    """
    Initialize sub-parser.
    Parameters: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser
    """
    return {
        "help": "Generates wikidata triples from kgtk file",
        "description": "Generating Wikidata triples.",
    }


def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
        prop_file: str, labelSet: str, aliasSet: str, descriptionSet: str, n: str, dest: Any  --output-n-lines --generate-truthy
    """
    from kgtk.utils.argparsehelpers import optional_bool

    parser.add_argument(
        "-lp",
        "--label-property",
        action="store",
        type=str,
        default="label",
        required=False,
        help="property identifiers which will create labels, separated by comma','.",
        dest="labels",
    )
    parser.add_argument(
        "-ap",
        "--alias-property",
        action="store",
        type=str,
        required=False,
        default="alias",
        help="alias identifiers which will create labels, separated by comma','.",
        dest="aliases",
    )
    parser.add_argument(
        "-dp",
        "--description-property",
        action="store",
        type=str,
        required=False,
        default="description",
        help="description identifiers which will create labels, separated by comma','.",
        dest="descriptions",
    )
    parser.add_argument(
        "-pf",
        "--property-file",
        action="store",
        type=str,
        required=False,
        default=None,
        help="path to the file which contains the property datatype mapping in kgtk format.",
        dest="property_file",
    )
    parser.add_argument(
        "-pd",
        "--property-declaration-in-file",
        dest="prop_declaration",
        metavar="True|False",
        help="whether read properties in the kgtk file. If set to yes, use `cat input.tsv input.tsv` to pipe the input file twice",
        type=optional_bool,
        nargs='?',
        const=True,
        default=False
    )
    parser.add_argument(
        "-n",
        "--output-n-lines",
        action="store",
        type=int,
        required=False,
        default=1000,
        help="output triples approximately every {n} lines of reading stdin.",
        dest="n",
    )
    parser.add_argument(
        "-gt",
        "--generate-truthy",
        type=optional_bool,
        metavar="True|False",
        nargs='?',
        const=True,
        default=True,
        help="the default is to not generate truthy triples. Specify this option to generate truthy triples.",
        dest="truthy",
    )
    parser.add_argument(
        "-w",
        "--warning",
        type=optional_bool,
        nargs='?',
        const=True,
        default=False,
        metavar="True|False",
        help="if specified, "
             "warn various kinds of exceptions and mistakes and log them to a log file with line number in input file, "
             "rather than stopping. logging",
        dest="warning",
    )

    parser.add_argument(
        "-sid",
        "--use-id",
        type=optional_bool,
        metavar="True|False",
        nargs='?',
        const=True,
        default=False,
        help="if set to yes, the id in the edge will be used as statement id when creating statement or truthy statement",
        dest="use_id",
    )
    parser.add_argument(
        "-log",
        "--log-path",
        action="store",
        type=str,
        required=False,
        default="warning.log",
        help="set the path of the log file",
        dest="log_path",
    )
    parser.add_argument(
        "-prefix",
        "--prefix-path",
        action="store",
        type=str,
        required=False,
        default="NONE",
        help="set the path of the prefix kgtk file that provides customized uri prefix binding",
        dest="prefix_path",
    )
    parser.add_argument(
        "--error-action",
        action="store",
        type=str,
        required=False,
        default="log",
        help="Defines the command behavior in case there are errors in execution, [log|raise]. "
             "'log': log the errors to a log file and continue,  'raise': raise exception and quit. Default: 'log'",
        dest="error_action",
    )
    parser.add_input_file(positional=True)
    parser.add_output_file(
        who="Output triples file path.",
        allow_list=False, dest="output_file")


def run(
        labels: str,
        aliases: str,
        descriptions: str,
        property_file: str,
        n: int,
        truthy: bool,
        warning: bool,
        use_id: bool,
        log_path: str,
        prop_declaration: bool,
        prefix_path: str,
        input_file: KGTKFiles,
        output_file: str,
        error_action: str
):
    # import modules locally

    from kgtk.generator import TripleGenerator
    from kgtk.exceptions import KGTKException

    generator = TripleGenerator(
        prop_file=property_file,
        label_set=labels,
        alias_set=aliases,
        description_set=descriptions,
        n=n,
        warning=warning,
        truthy=truthy,
        use_id=use_id,
        dest_fp=output_file,
        log_path=log_path,
        prop_declaration=prop_declaration,
        prefix_path=prefix_path,
        input_file=input_file,
        error_action=error_action
    )

    try:
        generator.process()
    except Exception as e:
        raise KGTKException(e)



"""This argparse type conversion function implements optional boolean arguments.

--arg
--arg=True
--arg=False

and other variations.  A default value of None is allowed for fallback
argument composition.

Sample usage:

parser.add_argument(prefix1 + "gzip-in-parallel",
                    dest=prefix2 + "gzip_in_parallel",
                    help=h(prefix3 + "Execute gzip in parallel (default=%(default)s)."),
                    metavar="True|False",
                    type=optional_bool, nargs='?', const=True, **d(default=False))

type=optional_bool      This says that the argument is an optional boolean.

nargs='?'               This says that the optional boolean can optionally take one
                        argument itself (True or False).

const=True              This says that if the optional boolean is present *without*
                        a following True or False, the value is True.

                        e.g.:
                        --gzip-in-parallel              means True
                        --gzip-in-parallel=True         means True
                        --gzip-in-parallel True         means True
                        --gzip-in-parallel=False        means False
                        --gzip-in-parallel False        means False
                        

default=False           This says that if the optional boolean isn't present at
                        all, the default value is True.

                        e.g.:
                        (--gzip-in-parallel isn't specified) means False

metavar="True|False"    The argument value are True or False (ignoring case).
                        Other values are also supported: f or t, 0 or 1, no or yes,
                        n or y.

In order for optional booleans to work, the following *must* be present:

nargs='?'
const=True
default=False

"""

import typing
from kgtk.exceptions import KGTKArgumentParseException


def optional_bool(value) -> typing.Optional[bool]:
    if value is None:
        return None
    if isinstance(value, bool):
        return value
    if value.lower() in {'false', 'f', '0', 'no', 'n'}:
        return False
    elif value.lower() in {'true', 't', '1', 'yes', 'y'}:
        return True
    raise KGTKArgumentParseException(f'{value} is not a valid boolean value')



"""
Read a KGTK node file in TSV format.

TODO: Add support for alternative envelope formats, such as JSON.
"""

from argparse import ArgumentParser
import attr
from pathlib import Path
import sys
import typing

from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.utils.closableiter import ClosableIter
from kgtk.utils.enumnameaction import EnumNameAction
from kgtk.utils.validationaction import ValidationAction
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=False)
class NodeReader(KgtkReader):

    @classmethod
    def open_node_file(cls,
                       file_path: typing.Optional[typing.Union[Path, str]],
                       who: str = "node input",
                       error_file: typing.TextIO = sys.stderr,
                       reject_file: typing.Optional[typing.TextIO] = None,
                       options: typing.Optional[KgtkReaderOptions] = None,
                       value_options: typing.Optional[KgtkValueOptions] = None,
                       verbose: bool = False,
                       very_verbose: bool = False)->"NodeReader":

        result: KgtkReader = cls.open(file_path=file_path,
                                      who=who,
                                      error_file=error_file,
                                      reject_file=reject_file,
                                      mode=KgtkReaderMode.NODE,
                                      options=options,
                                      value_options=value_options,
                                      verbose=verbose,
                                      very_verbose=very_verbose)
        # This doesn't work because the EdgeReader imported inside KgtkReader
        # is a different class than this one!
        #
        # TODO: Fix this.
        #
        #if isinstance(result, cls):
        #    return result
        #else:
        #    # TODO: throw a better exception
        #    raise ValueError("open_node_file expected to produce a NodeReader")
        return typing.cast(NodeReader, result)

    def _ignore_prohibited_lists(self, row: typing.List[str], line: str)->bool:
        # TODO: Should we apply this to the `id` column of a node file for simplicity?
        return False

    def _ignore_if_blank_required_fields(self, values: typing.List[str], line: str)->bool:
        # Ignore line_action with blank id fields.  This code comes after
        # filling missing trailing columns, although it could be reworked
        # to come first.
        if self.options.blank_required_field_line_action != ValidationAction.PASS and self.id_column_idx >= 0 and len(values) > self.id_column_idx:
            id_value: str = values[self.id_column_idx]
            if len(id_value) == 0 or id_value.isspace():
                return self.exclude_line(self.options.blank_required_field_line_action, "id is blank", line)
        return False # Do not ignore this line

    def _skip_reserved_fields(self, column_name)->bool:
        if self.id_column_idx >= 0 and column_name in self.ID_COLUMN_NAMES:
            return True
        return False

def main():
    """
    Test the KGTK node file reader.
    """
    parser = ArgumentParser()
    parser.add_argument(dest="kgtk_file", help="The KGTK edge file to read", type=Path, nargs="?")
    KgtkReader.add_debug_arguments(parser, expert=True)
    KgtkReaderOptions.add_arguments(parser, validate_by_default=True, expert=True)
    KgtkValueOptions.add_arguments(parser, expert=True)
    args = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args, mode=KgtkReaderMode.NODE)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

    if args.show_options:
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    nr: NodeReader = NodeReader.open_node_file(args.kgtk_file,
                                               error_file=error_file,
                                               options=reader_options,
                                               value_options=value_options,
                                               verbose=args.verbose, very_verbose=args.very_verbose)

    line_count: int = 0
    row: typing.List[str]
    for row in nr:
        line_count += 1
    print("Read %d lines" % line_count, file=error_file, flush=True)

if __name__ == "__main__":
    main()




import sys
from pathlib import Path
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
from kgtk.exceptions import KGTKException
from typing import Tuple, Optional

from kgtk.io.kgtkwriter import KgtkWriter

WORD2VEC = 'word2vec'
GPROJECTOR = 'gprojector'


class ConvertEmbeddingsFormat(object):
    def __init__(self,
                 input_file: Path,
                 output_file: Path,
                 node_file: Path = None,
                 output_format: Optional[str] = WORD2VEC,
                 input_property: Optional[str] = 'embeddings',
                 metadata_columns: Optional[str] = None,
                 output_metadata_file: Path = None,
                 error_file: str = sys.stderr,
                 line_separator: str = ",",
                 **kwargs
                 ):
        if output_format not in (WORD2VEC, GPROJECTOR):
            raise KGTKException(f'--output-format should be one of "{WORD2VEC}" or "{GPROJECTOR}')
        self.input_file = input_file
        self.output_file = output_file
        self.node_file = node_file
        self.output_format = output_format.lower()
        self.input_property = input_property
        self.line_separator = line_separator
        if output_metadata_file is None:
            output_metadata_file = f"{str(Path.home())}/kgtk_embeddings_gprojector_metadata.tsv"
        self.output_metadata_file = output_metadata_file

        self.metadata_columns = metadata_columns.split(",") if metadata_columns is not None else []

        self.error_file = error_file

        self.reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)

    def process(self):
        kw_metadata = None
        node_metadata = {}

        row_count = 0
        if self.output_format == GPROJECTOR:
            node_metadata = self.read_node_file()
            metadata_column_names = node_metadata['column_names'] if node_metadata else self.metadata_columns
            kw_metadata = KgtkWriter.open(file_path=self.output_metadata_file,
                                          error_file=self.error_file,
                                          column_names=metadata_column_names,
                                          mode=KgtkWriter.Mode.NONE,
                                          require_all_columns=False,
                                          no_header=False)

        kw: KgtkWriter = KgtkWriter.open(file_path=self.output_file,
                                         error_file=self.error_file,
                                         column_names=[],
                                         mode=KgtkWriter.Mode.NONE,
                                         require_all_columns=False,
                                         no_header=True)

        kr: KgtkReader = KgtkReader.open(self.input_file,
                                         error_file=self.error_file,
                                         mode=KgtkReaderMode.EDGE,
                                         options=self.reader_options
                                         )
        if self.node_file is None and self.output_format == GPROJECTOR and len(self.metadata_columns) > 0:
            columns_not_found = []
            for col in self.metadata_columns:
                if col not in kr.column_name_map:
                    columns_not_found.append(col)
            if len(columns_not_found) > 0:
                kr.close()
                kw.close()
                if kw_metadata:
                    kw_metadata.close()
                raise KGTKException(
                    f"The following metadata columns are not found in the input file: {','.join(columns_not_found)}")

        if self.output_format == WORD2VEC:
            # first line is number of vectors and vector dimension
            line_count, dimension_count = self.count_lines(kr)
            kw.writeline(f"{line_count} {dimension_count}")

            kr.close()

            kr: KgtkReader = KgtkReader.open(self.input_file,
                                             error_file=self.error_file,
                                             mode=KgtkReaderMode.EDGE,
                                             options=self.reader_options
                                             )

        for row in kr:
            node1 = row[kr.node1_column_idx]
            node2 = row[kr.node2_column_idx]
            if row[kr.label_column_idx] == self.input_property:
                if self.output_format == WORD2VEC:
                    kw.writeline(f"{node1} {self.line_to_w2v(node2)}")
                elif self.output_format == GPROJECTOR:
                    if row_count == 10000:  # google projector will only handle upto 10000 rows
                        break
                    row_count += 1
                    node2_gp = self.line_to_gprojector(node2)
                    node1_metadata = ""
                    if node2_gp != "":
                        kw.writeline(node2_gp)
                        if node_metadata:
                            if node1 in node_metadata:
                                node1_metadata = "\t".join(node_metadata[node1].values())
                        elif len(self.metadata_columns) > 0:
                            values = [row[kr.column_name_map[x]] for x in self.metadata_columns]
                            node1_metadata = "\t".join(values)
                        kw_metadata.writeline(node1_metadata)

        kw.close()
        kr.close()
        if kw_metadata:
            kw_metadata.close()

    def count_lines(self, kr: KgtkReader) -> Tuple[int, int]:
        line_count = 0
        dimension_count = 0
        for row in kr:
            node2 = row[kr.node2_column_idx]
            if row[kr.label_column_idx] == self.input_property:
                if dimension_count == 0:
                    dimension_count = len(node2.split(self.line_separator))
                line_count += 1
        if line_count == 0 and dimension_count == 0:
            raise KGTKException(f"Zero rows in the input file with property: {self.input_property}")

        return line_count, dimension_count

    def line_to_w2v(self, line: str) -> str:
        if line is None or line.strip() == "":
            return ""

        return " ".join(line.strip().split(self.line_separator))

    def line_to_gprojector(self, line: str) -> str:
        if line is None or line.strip() == "":
            return ""

        return "\t".join(line.strip().split(self.line_separator))

    def read_node_file(self) -> dict:
        node_metadata = {}
        if self.node_file is None:
            return node_metadata

        kr_node: KgtkReader = KgtkReader.open(self.node_file,
                                              error_file=self.error_file,
                                              options=self.reader_options,
                                              mode=KgtkReaderMode.NONE,
                                              )
        columns_not_found = []
        if len(self.metadata_columns) > 0:
            for col in self.metadata_columns:
                if col not in kr_node.column_names:
                    columns_not_found.append(col)
            node_metadata['column_names'] = self.metadata_columns
        else:
            node_metadata['column_names'] = kr_node.column_names

        if len(columns_not_found) > 0:
            kr_node.close()
            raise KGTKException(
                f"The following metadata columns are not found in the node file: {','.join(columns_not_found)}")

        for row in kr_node:
            if len(row) == len(kr_node.column_names):
                node_id = row[kr_node.id_column_idx]
                if node_id not in node_metadata:
                    node_metadata[node_id] = dict()
                for col in node_metadata['column_names']:
                    node_metadata[node_id][col] = row[kr_node.column_name_map[col]]
        kr_node.close()
        return node_metadata



from kgtk.knowledge_graph.subject import Subject
from kgtk.knowledge_graph.node import URI, Literal
from kgtk.wikidata.statement import Statement, Rank
from kgtk.wikidata.value import Item, Property, Datatype
from uuid import uuid4

change_recorder = set()
revision = None


def revise(enable=False):
    """
    Tracking revision
    :param enable: If False (default), tracking stops. If True, a new revision will be generated.
    """
    global revision

    if not enable:
        revision = None
    else:
        revision = URI('http://www.isi.edu/etk/revision#{}'.format(uuid4().hex))


def serialize_change_record(fp):
    fp.writelines('{}\t{}\n'.format(node, prop) for node, prop in change_recorder)


class Entity(Subject):
    def __init__(self, node, creator):
        super().__init__(URI('wd:' + node))
        self.node_id = node
        self.creator = URI(creator) if creator else None

    def add_label(self, s: str, lang='en'):
        literal = Literal(s, lang=lang)
        self.add_property(URI('rdfs:label'), literal)
        self.add_property(URI('schema:name'), literal)
        self.add_property(URI('skos:prefLabel'), literal)

    def add_alias(self, s: str, lang='en'):
        self.add_property(URI('skos:altLabel'), Literal(s, lang=lang))

    def add_description(self, s: str, lang='en'):
        self.add_property(URI('schema:description'), Literal(s, lang=lang))

    def add_statement(self, p: str, v, rank=Rank.Normal, statement_id=None):
        global revision

        change_recorder.add((self.node_id, p))
        statement = Statement(self.node_id, rank, statement_id=statement_id)
        statement.add_value(p, v)
        statement.add_property(URI('http://www.isi.edu/etk/createdBy'), self.creator)
        if revision:
            statement.add_property(URI('http://www.isi.edu/etk/revision'), revision)
        self.add_property(URI('p:' + p), statement)
        return statement

    def add_truthy_value_node(self, p, v):
        self.add_property(URI('wdt:' + p), v.value)
        if v.normalized_value:
            self.add_property(URI('wdtn:' + p), v.normalized_value)

    def add_truthy_statement(self, p: str, v, statement_id=None):
        statement = self.add_statement(p, v, Rank.BestRank, statement_id)

        self.add_truthy_value_node(p, v)
        return statement


class WDItem(Entity, Item):
    def __init__(self, s: str, creator='http://www.isi.edu/datamart'):
        Entity.__init__(self, s, creator)
        Item.__init__(self, s)
        self.add_property(URI('rdf:type'), URI('wikibase:Item'))


class WDProperty(Entity, Property):
    def __init__(self, s: str, property_type, creator='http://www.isi.edu/datamart'):
        Entity.__init__(self, s, creator)
        Property.__init__(self, s)
        self.add_property(URI('rdf:type'), URI('wikibase:Property'))
        type_uri = property_type if not isinstance(property_type, Datatype) else Datatype(property_type)
        self.add_property(URI('wikibase:propertyType'), type_uri.type)

        self.add_property(URI('wikibase:directClaim'), URI('wdt:' + s))
        self.add_property(URI('wikibase:directClaimNormalized'), URI('wdtn:' + s))
        self.add_property(URI('wikibase:claim'), URI('p:' + s))
        self.add_property(URI('wikibase:statementProperty'), URI('ps:' + s))
        self.add_property(URI('wikibase:statementValue'), URI('psv:' + s))
        self.add_property(URI('wikibase:statementValueNormalized'), URI('psn:' + s))
        self.add_property(URI('wikibase:qualifier'), URI('pq:' + s))
        self.add_property(URI('wikibase:qualifierValue'), URI('pqv:' + s))
        self.add_property(URI('wikibase:qualifierValueNormalized'), URI('pqn:' + s))
        self.add_property(URI('wikibase:reference'), URI('pr:' + s))
        self.add_property(URI('wikibase:referenceValue'), URI('prv:' + s))
        self.add_property(URI('wikibase:referenceValueNormalized'), URI('prn:' + s))
        self.add_property(URI('wikibase:novalue'), URI('wdno:' + s))



"""Copy records from the first KGTK file to the output file, if one or more
columns are (any/all) (not) empty.  If --only-count is True, report the count
of qualifying records but do not write the output file.

"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkIfEmpty(KgtkFormat):
    input_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    filter_column_names: typing.List[str] = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                            iterable_validator=attr.validators.instance_of(list)))

    output_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    reject_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)), default=None)

    all_are: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    notempty: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    only_count: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def filter(self,
               row: typing.List[str],
               filter_idx_list: typing.List[int])->bool:
        idx: int
        if self.notempty == False and self.all_are == False:
            # if any are empty.
            for idx in filter_idx_list:
                if len(row[idx]) == 0:
                    return True
            return False

        elif self.notempty == False and self.all_are == True:
            # if all are empty.
            for idx in filter_idx_list:
                if len(row[idx]) != 0:
                    return False
            return True
            
        elif self.notempty == True and self.all_are == False:
            # If any are not empty.
            for idx in filter_idx_list:
                if len(row[idx]) != 0:
                    return True
            return False
        
        else:
            # if all are not empty:
            for idx in filter_idx_list:
                if len(row[idx]) == 0:
                    return False
            return True

    def process(self):
        # Open the input file.
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        filter_idx_list: typing.List[int] = [ ]
        column_name: str
        for column_name in self.filter_column_names:
            if column_name not in kr.column_name_map:
                raise ValueError("Column %s is not in the input file" % (column_name))
            filter_idx_list.append(kr.column_name_map[column_name])
            

        ew: typing.Optional[KgtkWriter] = None
        rew: typing.Optional[KgtkWriter] = None
        if not self.only_count:
            if self.verbose:
                print("Opening the output file: %s" % self.output_file_path, file=self.error_file, flush=True)
            ew = KgtkWriter.open(kr.column_names,
                                 self.output_file_path,
                                 mode=kr.mode,
                                 require_all_columns=False,
                                 prohibit_extra_columns=True,
                                 fill_missing_columns=True,
                                 use_mgzip=self.reader_options.use_mgzip, # Hack!
                                 mgzip_threads=self.reader_options.mgzip_threads, # Hack!
                                 gzip_in_parallel=False,
                                 verbose=self.verbose,
                                 very_verbose=self.very_verbose)        
            
            if self.reject_file_path is not None:
                if self.verbose:
                    print("Opening the reject file: %s" % self.reject_file_path, file=self.error_file, flush=True)
                rew = KgtkWriter.open(kr.column_names,
                                      self.reject_file_path,
                                      mode=KgtkWriter.Mode[kr.mode.name],
                                      require_all_columns=False,
                                      prohibit_extra_columns=True,
                                      fill_missing_columns=True,
                                      use_mgzip=self.reader_options.use_mgzip, # Hack!
                                      mgzip_threads=self.reader_options.mgzip_threads, # Hack!
                                      gzip_in_parallel=False,
                                      verbose=self.verbose,
                                      very_verbose=self.very_verbose)
            
        if self.verbose:
            print("Filtering records from %s" % self.input_file_path, file=self.error_file, flush=True)
        input_line_count: int = 0
        output_line_count: int = 0;
        reject_line_count: int = 0;

        row: typing.list[str]
        for row in kr:
            input_line_count += 1
            if self.filter(row, filter_idx_list):
                if not self.only_count:
                    ew.write(row)
                output_line_count += 1
            else:
                if not self.only_count and rew is not None:
                    rew.write(row)
                reject_line_count += 1


        if self.only_count:
            print("Read %d records, %d records passed the filter, %d rejected." % (input_line_count,
                                                                                   output_line_count,
                                                                                   reject_line_count),
                  file=self.error_file, flush=True)
        else:
            if self.verbose:
                if rew is not None:
                    print("Read %d records, wrote %d records, rejected %d records." % (input_line_count,
                                                                                       output_line_count,
                                                                                       reject_line_count),
                          file=self.error_file, flush=True)
                else:
                    print("Read %d records, wrote %d records." % (input_line_count, output_line_count), file=self.error_file, flush=True)
                    
        
            ew.close()
            if rew is not None:
                rew.close()

def main():
    """
    Test the KGTK ifempty processor.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data", type=Path, nargs="?")

    parser.add_argument(      "--columns", dest="filter_column_names", help="The columns to filter on (default=None).", nargs='+', required=True)

    parser.add_argument(      "--count", dest="only_count", help="Only count the records, do not copy them. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--all", dest="all_are", help="False: Test if any are, True: test if all are (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--not-empty", dest="notempty", help="False: test if empty, True: test if not empty (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)


    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        # TODO: show ifempty-specific options.
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    ie: KgtkIfEmpty = KgtkIfEmpty(
        input_file_path=args.input_file_path,
        filter_column_names=args.filter_column_names,
        output_file_path=args.output_file_path,
        all_are=args.all_are,
        notempty=args.notempty,
        only_count = args.only_count,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    ie.process()

if __name__ == "__main__":
    main()



from pyshacl import validate
from typing import Dict, Union
from kgtk.knowledge_graph.graph import Graph
from kgtk.knowledge_graph.node import URI, BNode, Literal, LiteralType
from kgtk.knowledge_graph.subject import Subject
from rdflib import RDF, RDFS, Namespace, XSD
import rdflib

SH = Namespace('http://www.w3.org/ns/shacl#')


class SHACL(Graph):
    def __init__(self):
        super().__init__()
        self._class_nodes: Dict[str, Union[URI, BNode]] = {}
        self.bind('sh', SH)

    def add_ontology(self, onto_graph):
        converter = SHACLOntoConverter(self._class_nodes)
        self._g += converter.convert_ontology(onto_graph)._g
        self._class_nodes.update(converter.class_nodes)

    def validate(self, data_graph, onto_graph=None, inference=None):
        if isinstance(data_graph, Graph):
            data_graph = data_graph._g
        if isinstance(onto_graph, Graph):
            onto_graph = onto_graph._g
        if onto_graph and inference:
            conforms, results_graph, results_text = validate(data_graph + onto_graph, shacl_graph=self._g,
                                                             inference=inference)
        else:
            conforms, results_graph, results_text = validate(data_graph, shacl_graph=self._g)
        results_graph = Graph(results_graph)
        return conforms, results_graph


class SHACLOntoConverter:
    def __init__(self, class_nodes: Dict[str, Union[URI, BNode]] = None):
        super().__init__()
        self.onto_graph: rdflib.Graph = None
        self._g = Graph()
        self._class_nodes = class_nodes if class_nodes else {}
        self._g.bind('sh', SH)

    def cnode(self, class_):
        class_ = str(class_)
        if class_ not in self._class_nodes:
            self._class_nodes[class_] = BNode()
        return self._class_nodes[class_]

    @property
    def class_nodes(self):
        return self._class_nodes

    def convert_ontology(self, onto_graph: Graph) -> Graph:
        self.onto_graph = onto_graph._g if isinstance(onto_graph, Graph) else onto_graph
        # build shacl property-based NodeShape for non-domain-referenced properties
        self._convert_ontology_non_referenced_property()
        # build shacl class-based NodeShape
        # build shacl property path-based blank node for domain-referenced properties
        # for path-based property,
        # we can only add sh:datatype/sh:class based on rdfs:range
        self._convert_ontology_class_node_shape()

        # build shacl property based on owl restriction
        # we can add cardinality to it
        self._convert_ontology_owl_restriction()
        return self._g

    def _property_shape(self, property_: Union[rdflib.URIRef, str, URI]):
        if isinstance(property_, URI):
            property_ = property_.value
        property_ = rdflib.URIRef(property_)
        p_shape = Subject(BNode())
        p_shape.add_property(URI('sh:path'), URI(property_))
        ranges = list(self.onto_graph.objects(property_, RDFS.range))
        self._build_property_ranges(p_shape, ranges)
        return p_shape

    def _build_property_ranges(self, p_shape, ranges):
        if not ranges:
            return
        if len(ranges) == 1:
            range_ = ranges[0]
            type_ = URI('sh:datatype') if self._is_datatype(range_) else URI('sh:class')
            p_shape.add_property(type_, URI(ranges[0]))
        else:
            or_list = []
            for range_ in ranges:
                range_subject = Subject(BNode())
                type_ = URI('sh:datatype') if self._is_datatype(range_) else URI('sh:class')
                range_subject.add_property(type_, URI(range_))
                or_list.append(range_subject)
            p_shape.add_property(URI('sh:or'), self._add_list(or_list))

    @staticmethod
    def _is_datatype(uri: rdflib.URIRef):
        if isinstance(uri, rdflib.BNode):
            return False
        return uri.startswith(str(XSD)) or uri.startswith(str(RDF))

    def _convert_ontology_non_referenced_property(self):
        for p, in self.onto_graph.query("""
          SELECT ?p
          WHERE {
            { ?p a rdfs:Property }
            UNION
            { ?p a owl:ObjectProperty }
            UNION
            { ?p a owl:DatatypeProperty }
            FILTER NOT EXISTS { ?p rdfs:domain ?c }
            FILTER NOT EXISTS { ?c owl:onProperty ?p }
          }
        """):
            p_shape = self._property_shape(p)
            p_shape.add_property(URI('rdf:type'), URI('sh:PropertyShape'))
            self._g.add_subject(p_shape)

    def _convert_ontology_class_node_shape(self):
        for c, in self.onto_graph.query("""
          SELECT DISTINCT ?s WHERE {
            {?s a owl:Class}
            UNION
            {?s a rdfs:Class}
            UNION
            {?p rdfs:domain ?s}
            UNION
            {?s rdfs:subClassOf|owl:equivalentClass ?o}
            UNION
            {?os rdfs:subClassOf|owl:equivalentClass ?s FILTER NOT EXISTS {?s a owl:Restriction}}
            FILTER isURI(?s)
          }
        """):
            """
            [ a sh:NodeShape ;
              sh:targetClass ?s ;
              sh:property [ sh:path p ; sh:class class ;] ;]
            """
            node_subject = self._class_shape(c)
            for p in self.onto_graph.subjects(RDFS.domain, c):
                property_subject = self._property_shape(p)
                node_subject.add_property(URI('sh:property'), property_subject)
            self._g.add_subject(node_subject)

    def _class_shape(self, c):
        node_subject = Subject(self.cnode(c))
        node_subject.add_property(URI('rdf:type'), URI('sh:NodeShape'))
        node_subject.add_property(URI('sh:targetClass'), URI(c))
        return node_subject

    def _convert_ontology_owl_restriction(self):
        for c, p, exact, min_, max_ in self.onto_graph.query("""
          SELECT ?c ?p ?exact ?min ?max
          WHERE {
            ?c rdfs:subClassOf|owl:equivalentClass ?res .
            ?res a owl:Restriction ;
                 owl:onProperty ?p .
            OPTIONAL {?res owl:cardinality ?exact}
            OPTIONAL {?res owl:minCardinality ?min}
            OPTIONAL {?res owl:maxCardinality ?max}
          }
        """):
            node_subject = self._class_shape(c)
            property_shape = self._property_shape(p)
            if exact:
                property_shape.add_property(URI('sh:count'), Literal(str(exact), type_=LiteralType.integer))
            if min_:
                property_shape.add_property(URI('sh:minCount'), Literal(str(min_), type_=LiteralType.integer))
            if max_:
                property_shape.add_property(URI('sh:maxCount'), Literal(str(max_), type_=LiteralType.integer))
            ranges = []
            for r, in self.onto_graph.query("""
              SELECT ?r
              WHERE {
                ?c rdfs:subClassOf|owl:equivalentClass ?res .
                ?res a owl:Restriction ;
                     owl:onProperty ?p ;
                     owl:allValuesFrom|owl:someValuesFrom|owl:hasValue ?r
              }
            """, initBindings={'c': c, 'p': p}):
                ranges.append(r)
            self._build_property_ranges(property_shape, ranges)
            node_subject.add_property(URI('sh:property'), property_shape)
            self._g.add_subject(node_subject)

    def _add_list(self, list_):
        """
        (:a :b :c) ===
        [ rdf:first :a ; rdf:rest [ rdf:first :b ; rdf:rest [ rdf:first :c ; rdf:rest rdf:nil]]]
        """
        if not list_:
            return URI('rdf:nil')
        head = Subject(BNode())
        head.add_property(URI('rdf:first'), list_[0])
        head.add_property(URI('rdf:rest'), self._add_list(list_[1:]))
        return head



"""
Read a KGTK file in TSV format using the fast path.
"""

import attr
import typing

from kgtk.io.kgtkreader import KgtkReader

@attr.s(slots=True, frozen=False)
class FastReader(KgtkReader):
# This is the fast read path.

    # Get the next edge values as a list of strings.
    def nextrow(self)-> typing.List[str]:
        line: str
        row: typing.List[str]

        # Read a line from the source
        try:
            line = next(self.source) # Will throw StopIteration when done.

        except StopIteration as e:
            # Close the input file!
            #
            # TODO: implement a close() routine and/or whatever it takes to support "with".
            self.source.close() # Do we need to guard against repeating this call?
            raise e

        # Count the data line read and passed on:
        self.data_lines_read += 1
        self.data_lines_passed += 1

        # Strip the end-of-line characters and split the fields:
        return line.rstrip("\r\n").split(self.options.column_separator)

@attr.s(slots=True, frozen=False)
class FilteredFastReader(KgtkReader):
# This is the fast read path with a filter.

    # Get the next edge values as a list of strings.
    def nextrow(self)-> typing.List[str]:
        line: str
        row: typing.List[str]

        while True:
            # Read a line from the source
            try:
                line = next(self.source) # Will throw StopIteration when done.

            except StopIteration as e:
                # Close the input file!
                #
                # TODO: implement a close() routine and/or whatever it takes to support "with".
                self.source.close() # Do we need to guard against repeating this call?
                raise e

            # Count the data line read and passed on:
            self.data_lines_read += 1

            if self.input_filter is not None:
                input_filter_idx: int
                input_filter_set: typing.Set[str]
                fail: bool = False
                for input_filter_idx, input_filter_set in self.input_filter.items():
                    value: str = row[input_filter_idx]
                    if value not in input_filter_set:
                        fail = True
                        break
                if fail:
                    self.data_lines_excluded_by_filter += 1
                    continue

            self.data_lines_passed += 1

            # Strip the end-of-line characters and split the fields:
            return line.rstrip("\r\n").split(self.options.column_separator)



from argparse import Namespace
import typing
from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
import sys

ALL_EMBEDDING_MODELS_NAMES = [
    "bert-base-nli-cls-token",
    "bert-base-nli-max-tokens",
    "bert-base-nli-mean-tokens",
    "bert-base-nli-stsb-mean-tokens",
    "bert-base-wikipedia-sections-mean-tokens",
    "bert-large-nli-cls-token",
    "bert-large-nli-max-tokens",
    "bert-large-nli-mean-tokens",
    "bert-large-nli-stsb-mean-tokens",
    "distilbert-base-nli-mean-tokens",
    "distilbert-base-nli-stsb-mean-tokens",
    "distiluse-base-multilingual-cased",
    "roberta-base-nli-mean-tokens",
    "roberta-base-nli-stsb-mean-tokens",
    "roberta-large-nli-mean-tokens",
    "roberta-large-nli-stsb-mean-tokens",
    "sentence-transformers/all-distilroberta-v1"
]


def parser():
    return {
        'help': """Produce embedding vectors on given file's nodes."""
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    _expert: bool = parsed_shared_args._expert

    parser.accept_shared_argument('_debug')

    parser.add_input_file(positional=True)

    # model name
    all_models_names = ALL_EMBEDDING_MODELS_NAMES
    parser.add_argument('-m', '--model', action='store', nargs='+', dest='all_models_names',
                        default="bert-base-nli-cls-token", choices=all_models_names,
                        help="the model to used for embedding")

    parser.add_argument('--sentence-property', action='store',
                        dest='sentence_property', default="sentence",
                        help="""The name of the property with sentence for each Qnode. Default is 'sentence'""")

    parser.add_argument('--output-property', action='store',
                        dest='output_properties', default="text_embedding",
                        help="""The output property name used to record the embedding. Default is `output_properties`.
                        This argument is only valid for output in kgtk format.""")

    parser.add_argument('--batch-size', action='store',
                        dest='batch_size', default=100000, type=int,
                        help="""The number of sentences to be processed at a time. Default is 100000.
                        Set this value to '-1' to process the whole file as one batch""")

    parser.add_argument('-o', '--out-file', action='store', dest='output_file', default="-",
                        help="output path for the text embedding file, by default it will be printed in console")

    parser.add_argument('--output-data-format', action='store', dest='output_data_format',
                        default="kgtk", choices=("w2v", "kgtk"),
                        help="output format, can either be `w2v` or `kgtk`. \n"
                             "If choose `w2v`, the output will be a text file, with each row contains the qnode and "
                             " the vector representation, separated by a space. The first line is the number of qnodes"
                             " and dimension of vectors, separated by space")

    KgtkReader.add_debug_arguments(parser, expert=False)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=False)


def run(**kwargs):
    from kgtk.exceptions import KGTKException
    import logging
    import os
    from time import strftime
    do_logging = kwargs.get("_debug", False)
    if do_logging:
        logging_level_class = logging.DEBUG
        logger_path = os.path.join(os.environ.get("HOME"),
                                   "kgtk_text_embedding_log_{}.log".format(strftime("%Y-%m-%d-%H-%M")))
        logging.basicConfig(level=logging_level_class,
                            format="%(asctime)s [%(levelname)s] %(name)s %(lineno)d -- %(message)s",
                            datefmt='%m-%d %H:%M:%S',
                            filename=logger_path,
                            filemode='w')

    _logger = logging.getLogger(__name__)
    _logger.warning("Running with logging level {}".format(_logger.getEffectiveLevel()))

    try:
        from pathlib import Path
        from kgtk.gt.embedding_utils import EmbeddingVector

        all_models_names = kwargs.get("all_models_names", ['bert-base-wikipedia-sections-mean-tokens'])
        output_format = kwargs.get("output_data_format")

        output_file = kwargs.get("output_file")
        batch_size = kwargs.get("batch_size")
        sentence_property = kwargs.get("sentence_property")
        output_properties = kwargs.get("output_properties")

        # Select where to send error messages, defaulting to stderr.
        error_file: typing.TextIO = sys.stdout if kwargs.get("errors_to_stdout") else sys.stderr

        # Build the option structures.
        reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
        value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

        verbose: bool = kwargs.get("verbose")

        input_file_path: Path = KGTKArgumentParser.get_input_file(kwargs.get("input_file"))

        if isinstance(all_models_names, str):
            all_models_names = [all_models_names]

        if len(all_models_names) == 0:
            raise KGTKException("No embedding vector model name given!")

        for each_model_name in all_models_names:
            _logger.info("Running {} model on {}".format(each_model_name, str(input_file_path)))
            process = EmbeddingVector(each_model_name,
                                      output_property_name=output_properties,
                                      sentence_property_name=sentence_property,
                                      output_format=output_format)

            process.process_sentences_kgtk(input_file_path=input_file_path,
                                           output_file_path=output_file,
                                           error_file=error_file,
                                           reader_options=reader_options,
                                           value_options=value_options,
                                           verbose=verbose,
                                           batch_size=batch_size
                                           )

            _logger.info("*" * 20 + "finished" + "*" * 20)
    except Exception as e:
        _logger.debug(e, exc_info=True)
        raise KGTKException(str(e))
